import logging
import math
import re
import io
from typing import List, Dict, Tuple

import numpy as np
import pandas as pd
from scipy import optimize
from scipy.interpolate import interp1d


def _norm_space(s):
    s -= np.min(s)
    return s / np.max(s)


def _cubic_polynomial(x, y, p):
    """Cubic polynomial that can be fitted to an isochrone.

    Args:
        x: Colours of points in isochrone.
        y: Magnitudes of points in isochrone.
        p: Coefficients for polynomial.

    Returns:
        Evaluated polynomial.
    """
    return p[0] + p[1] * x + p[2] * y + p[3] * x * x + p[4] * y * y + p[5] * x * y + p[6] * x * x * x + \
           p[7] * x * x * y + p[8] * x * y * y + p[9] * y * y * y


def _quadratic_polynomial(x, y, p):
    """Quadratic polynomial that can be fitted to an isochrone.

    Args:
        x: Colours of points in isochrone.
        y: Magnitudes of points in isochrone.
        p: Coefficients for polynomial.

    Returns:
        Evaluated polynomial.
    """
    return p[0] + p[1] * x + p[2] * y + p[3] * x * x + p[4] * y * y + p[5] * x * y


def _linear_polynomial(x, y, p):
    """Linear polynomial that can be fitted to an isochrone.

    Args:
        x: Colours of points in isochrone.
        y: Magnitudes of points in isochrone.
        p: Coefficients for polynomial.

    Returns:
        Evaluated polynomial.
    """
    return p[0] + p[1] * x + p[2] * y


POLYNOMIALS = {
    'linear': _linear_polynomial,
    'quadratic': _quadratic_polynomial,
    'cubic': _cubic_polynomial
}


class Isochrone:
    """Handles an isochrone in spexxy."""
    def __init__(self, filename: str = None, data: pd.DataFrame = None, meta: dict = None):
        """Create a new isochrone.
        
        Args:
            data: Actual isochrone data with column Teff, logg, logL/Lo, mbol, M_ini, M_act, and one for each filter.
            meta: Meta data for the isochrone, i.e. age, metallicity, etc.
        """

        # what do we have?
        if filename is not None:
            # read from file, so try to read meta data
            self._meta = {}
            with open(filename, 'r') as f:
                for line in f:
                    # we break on first non-comment line
                    if not line.startswith('#'):
                        break

                    # split by first '=', if exists
                    eq = line.find('=')
                    if eq > -1:
                        # get it
                        key = line[1:eq].strip()
                        val = line[eq + 1:].strip()

                        # is val a float?
                        try:
                            val = float(val)
                        except ValueError:
                            pass

                        # store it
                        self._meta[key] = val
            # load data
            self._data = pd.read_csv(filename, index_col=False, comment='#')

        elif data is not None:
            # take data and meta
            self._data = data
            self._meta = meta

        else:
            raise ValueError('Either filename or data must be given.')

    @staticmethod
    def load(filename: str):
        return Isochrone(filename)

    @property
    def age(self):
        return self._meta['Age'] if 'Age' in self._meta else None

    @property
    def metallicity(self):
        return self._meta['[M/H]'] if '[M/H]' in self._meta else None

    @property
    def extinction(self):
        return self._meta['Av'] if 'Av' in self._meta else None

    @property
    def type(self):
        return self._meta['Type'] if 'Type' in self._meta else None

    @property
    def filters(self) -> list:
        """Returns a list of filters defines by this isochrone.

        Returns:
            List of filters.
        """
        return list(self._data.columns.values[6:])

    @property
    def data(self) -> pd.DataFrame:
        """Returns the full isochrone data.

        Returns:
            Isochrone data table.
        """
        return self._data

    def __getitem__(self, item: str) -> float:
        """Returns a meta data item.

        Args:
            item: Name of meta data.

        Returns:
            Meta data value.
        """
        return self._meta[item] if item in self._meta else None

    def __setitem__(self, item: str, value: float):
        """Sets a given meta data item.

        Args:
            item: Name of meta data.
            value: New value for meta data.
        """
        self._meta[item] = value

    def copy(self) -> 'Isochrone':
        """Create a deep copy of this isochrone.

        Returns:
            Copy of this isochrone.
        """
        return Isochrone(data=self._data.copy(), meta=dict(self._meta))

    def save(self, filename: str):
        """Write isochrone to a file.

        Args:
            filename: Name of file to write isochrone in.
        """

        # open file
        with open(filename, 'w') as f:
            # write all meta data
            for key, val in self._meta.items():
                f.write('# %s=%s\n' % (key, str(val)))

            # write data (for some reason this StringIO workaround seems to be necessary sometimes...
            with io.StringIO() as sio:
                self._data.to_csv(sio, index=False)
                f.write(sio.getvalue())

    def _fit(self, col: list, mag: list, value: list, mode: str = 'linear'):
        """Fit a polynomial to the isochrone.

        Args:
            col: List (or equivalent) of colours to fit.
            mag: List (or equivalent) of magnitudes to fit.
            value: List (or equivalent) of data to fit.
            mode: Polynomial degree to use, either linear, quadratic, or cubic

        Returns:
            Result
        """

        # get polynomial
        poly = POLYNOMIALS[mode]

        # error function
        func = lambda p: poly(col, mag, p) - value

        # initial guess
        num_coeffs = {'linear': 3, 'quadratic': 6, 'cubic': 10}[mode] - 1
        initial = np.array([1] + [0] * num_coeffs)

        # fit
        res = optimize.leastsq(func, initial, full_output=True)
        if res[4] not in [1, 2, 3, 4]:
            raise ValueError("Could not fit isochrone: " + res[3])

        # return parameters
        return res[0]

    def interpolator(self, column: str, filter1: str, filter2: str, mode: str = 'linear'):
        """Create a polynomial interpolator for the given data column and two filters.

        Args:
            column: Name of column containing data.
            filter1: Name of first filter, used for magnitude.
            filter2: Name of second filter, colour is calculated as filter1-filter2.
            mode: Polynomial degree to use, either linear, quadratic, or cubic

        Returns:
            Function for interpolating isochrone with a polynomial.
        """

        # create colour column and get mag
        colour = self._data[filter1] - self._data[filter2]
        magnitudes = self._data[filter1]

        # get polynomial
        poly = POLYNOMIALS[mode]

        # fit polynomial
        coeffs = self._fit(colour, magnitudes, self._data[column], mode=mode)

        # define method to return
        def interpolator_inner(col, mag):
            return poly(col, mag, coeffs)

        # return method
        return interpolator_inner

    def nearest(self, filter1: str, filter2: str):
        """Creates an 'interpolator' that returns the nearest point on the isochrone for a colour/magnitude pair.

        Args:
            filter1: Name of first filter, used for magnitude.
            filter2: Name of second filter, colour is calculated as filter1-filter2.

        Returns:
            Function that calculates nearest neighbours on isochrone.
        """

        # create colour column and get mag
        colour = self._data[filter1] - self._data[filter2]
        magnitudes = self._data[filter1]

        # define method to return
        def nearest_inner(col, mag):
            # check
            if col is None or mag is None:
                return None

            # get nearest, scale colour with factor 7
            tmp = 49. * (colour - col)**2. + (magnitudes - mag)**2.
            d = np.argmin(tmp)

            # return values
            return {
                'dist': math.sqrt(tmp[d]),
                'v-i': self._data.loc[d, filter1] - self._data.loc[d, filter2],
                'v': self._data.loc[d, filter1],
                'teff': self._data.loc[d, 'Teff'],
                'logg': self._data.loc[d, 'logg'],
                'mini': self._data.loc[d, 'M_ini'],
                'mact': self._data.loc[d, 'M_act']
            }

        # return inner method
        return nearest_inner

    def apply_distance(self, distance: float):
        """Apply a distance in pc to the isochrone.

        Args:
            distance: Distance in pc.
        """

        # calculating distance modulus
        dist_mod = 5. * np.log10(distance) - 5.

        # set distance in meta
        self._meta['Distance'] = distance

        # add to filter columns
        self._data[self.filters] += dist_mod

    def keep_filters(self, filters: list):
        """Drops all filters except the given ones.
        
        Args:
            filters: List of filters to keep.
        """

        # loop all filters
        for f in list(self.filters):
            if f not in filters:
                self._data.drop(f, axis=1, inplace=True)

    def cut_regions(self, regions: list):
        """Tries to find regions in isochrone and cuts it down to the defined regions.

        Args:
            regions: List of region names (MS, SGB, RGB, HB).
        """

        # sort isochrone by M_ini
        isochrone = self._data[:]
        isochrone.sort_values('M_ini', inplace=True)

        # find MS by finding first turn-around for Teff
        Teff = isochrone['Teff']
        MS = None
        for i in range(len(Teff) - 1):
            if Teff.iloc[i + 1] < Teff.iloc[i]:
                logging.info('Found main sequence turnoff at %.2fK.', Teff.iloc[i])
                MS = isochrone.iloc[:i + 1]
                isochrone = isochrone.iloc[i + 1:]
                break

        # find tip of RGB by finding first turn-off for Teff at logg<1
        GB = None
        logg = isochrone['logg']
        Teff = isochrone['Teff']
        for i in range(len(Teff) - 1):
            if logg.iloc[i] < 1 and Teff.iloc[i + 1] > Teff.iloc[i]:
                logging.info('Found tip of RPG at %.2fK.', Teff.iloc[i])
                GB = isochrone.iloc[:i + 1]
                isochrone = isochrone.iloc[i + 1:]
                break

        # find HB by finding highest temperature
        Teff = isochrone['Teff']
        i = np.argmax(Teff.values)
        logging.info('Found end of horizontal branch at %.2fK.', Teff.iloc[i])
        HB = isochrone.iloc[i:]

        # connect parts again
        parts = []
        if 'MS' in regions:
            parts += [MS]
        if 'SGB' in regions or 'RGB' in regions:
            parts += [GB]
        if 'HB' in regions:
            parts += [HB]
        # if 'AGB' in regions:
        #    parts += [AGB]
        self._data = pd.concat(parts)

    def interpolate(self, count: int = 500, column: str = 'M_ini', space: list = None):
        """Interpolate an isochrone.

        Args:
            count: Number of new points along isochrone.
            column: Column used for interpolation.
            space: Two columns defining space for distance calculations.

        """

        # space
        if space is None:
            space = ['Teff', 'logg']

        # get X column for interpolator
        x = self.data[column]
        x_min = np.min(x)
        x_max = np.max(x)

        # normalize space to 0..1
        space_x = _norm_space(self.data[space[0]].values.copy())
        space_y = _norm_space(self.data[space[1]].values.copy())

        # integrate space along isochrone
        logging.info('Creating new grid with %d points in %s between %.3f and %.3f', count, column, x_min, x_max)
        dist = [0.]
        for i in range(len(space_x) - 1):
            dist += [dist[-1] + np.sqrt((space_x[i + 1] - space_x[i]) ** 2. + (space_y[i + 1] - space_y[i]) ** 2.)]

        # create interpolator of x on dist
        ip = interp1d(dist, x)

        # interpolate x with constant sampling, which will be our new grid
        grid = ip(np.linspace(0., dist[-1], count))

        # get new grid
        data = {column: grid}

        # interpolate
        for col in self.data.columns:
            # skip column from command line
            if col == column:
                continue
            logging.info('Interpolating column "%s"...', col)

            # create interpolator
            ip = interp1d(x.values, self.data[col].values)

            # do interpolation
            data[col] = ip(grid)

        # use it
        self._data = pd.DataFrame(data)


class MultiIsochrone:
    def __init__(self, filename: str, *args, **kwargs):
        # init
        self.filename = filename
        self._ages = []
        self._metallicities = []
        self._extinctions = []

        # load isochrones
        self._isochrones = self._load_isochrones(filename)

        # get parameters
        params = list(zip(*list(self._isochrones.keys())))
        self._ages = list(sorted(list(set(params[0]))))
        self._metallicities = list(sorted(list(set(params[1]))))
        self._extinctions = list(sorted(list(set(params[2]))))

    @staticmethod
    def guess_type(filename: str) -> str:
        """Guess type of given file.

        Args:
            filename: Name of file to guess type for.

        Returns:
            Type of file.
        """

        # open file
        with open(filename, 'r') as f:
            # read first line
            line = f.readline()

            # CMD?
            if 'CMD 2.7' in line:
                return 'CMD 2.7'
            elif 'CMD 2.9' in line:
                return 'CMD 2.9'
            elif 'CMD 3.3' in line:
                return 'CMD 3.3'
            else:
                raise ValueError('Unknown isochrone type.')

    @staticmethod
    def load(filename: str) -> 'MultiIsochrone':
        """Guess isochrone type and load it

        Args:
            filename: Name of file to load.

        Returns:
            A MultiIsochrone object.
        """

        # guess type
        file_type = MultiIsochrone.guess_type(filename)

        # load it
        if file_type == 'CMD 2.7':
            return CMD27IsochroneFile(filename)
        elif file_type == 'CMD 2.9':
            return CMD29IsochroneFile(filename)
        elif file_type == 'CMD 3.3':
            return CMD33IsochroneFile(filename)
        else:
            raise ValueError('Unknown isochrone type.')

    def _load_isochrones(self, filename: str) -> Dict[Tuple[float, float, float], Isochrone]:
        """Actually load isochrones from file.

        Args:
            filename: Name of file to load.

        Returns:
            Dictionary with isochrones.
        """
        raise NotImplemented

    @property
    def ages(self):
        return self._ages

    @property
    def metallicities(self):
        return self._metallicities

    @property
    def extinction(self):
        return self._extinctions

    def isochrone(self, age: float, metallicity: float, extinction: float) -> Isochrone:
        """Return a single isochrone.

        Args:
            age: Age
            metallicity: Metallicity [M/H]
            extinction: Extinction Av

        Returns:
            A single isochrone with the given parameters
        """
        return self._isochrones[(age, metallicity, extinction)]

    @property
    def isochrones(self) -> Dict[Tuple[float, float, float], Isochrone]:
        """Returns all isochrones in file.

        Returns:
            Dictionary with all Isochrones
        """
        return self._isochrones


class CMD27IsochroneFile(MultiIsochrone):
    """Isochrone file from PARSEC 2.7"""

    def _load_isochrones(self, filename: str) -> Dict[Tuple[float, float, float], Isochrone]:
        """Actually load isochrones from file.

        Args:
            filename: Name of file to load.

        Returns:
            Dictionary wuth isochrones.
        """

        # regular expression for search for end of header
        re_hdr = re.compile(r'\[M/H\]\s+=\s+([+-]?[0-9]+\.[0-9]+).*Age\s+=\s+([0-9]*\.[0-9]*e[+-][0-9]*)\s+yr')

        # find line with header
        header_lines = None
        m_h = None
        age = None
        header = None
        logging.info('Searching for header...')
        with open(filename, "r") as f:
            for header_lines, line in enumerate(f):
                m = re_hdr.search(line)
                if m:
                    m_h = float(m.group(1))
                    age = float(m.group(2))

                if line[0] == '#' and 'log(age/yr)' in line:
                    # found last header line, read header
                    header = line[1:].split()
                    break

        # no header or [M/H] found?
        if header is None:
            raise IOError('Could not find header.')
        if m_h is None or age is None:
            raise IOError('Could not find [M/H] or age.')
        logging.info('Found Age=%.2fgyr, [M/H]=%.2f.', age/1e9, m_h)

        # read file
        logging.info('Reading data...')
        data = pd.read_csv(filename, skiprows=header_lines + 1, sep='\s+', names=header)

        # find filter columns
        filters = list(data.columns.values[8:-2])
        logging.info('Found filters: ' + ', '.join(filters))

        # we don't want log(Teff), but Teff
        data['Teff'] = 10. ** data['logTe']

        # get meta data
        meta = {
            'Age': age,
            '[M/H]': m_h,
            'Type': 'CMD 2.7'
        }

        # rename some columns
        data = data.rename(columns={'logG': 'logg'})

        # log it
        logging.info('Found columns: ' + ', '.join(data.columns))

        # return isochrone
        iso = Isochrone(data=data[['Teff', 'logg', 'logL/Lo', 'mbol', 'M_ini', 'M_act'] + filters], meta=meta)
        return {(age, m_h, 0): iso}


class CMD29IsochroneFile(MultiIsochrone):
    """Isochrone file from PARSEC 2.9"""

    def _load_isochrones(self, filename: str) -> Dict[Tuple[float, float, float], Isochrone]:
        """Actually load isochrones from file.

        Args:
            filename: Name of file to load.

        Returns:
            Dictionary wuth isochrones.
        """

        # find line with header
        header = None
        logging.info('Searching for header...')
        with open(filename, "r") as f:
            last_line = None
            for line in f:
                # first line with no comment? means last line was header.
                if not line.startswith('#'):
                    header = last_line[1:].split()
                    break

                # store last line
                last_line = line

        # no header found?
        if header is None:
            raise IOError('Could not find header.')

        # read file
        logging.info('Reading data...')
        data = pd.read_csv(filename, comment='#', sep='\s+', names=header)

        # add metallicity column
        data['FeH'] = np.log(data['Zini'] / 0.0152)

        # remove "*mag" from columns
        data.columns = [c[:-3] if c.endswith('mag') else c for c in data.columns]

        # find filter columns
        filters = list(data.columns.values[24:])
        logging.info('Found filters: ' + ', '.join(filters))

        # we don't want log(Teff), but Teff
        data['Teff'] = 10. ** data['logTe']

        # rename some columns
        data = data.rename(columns={'Mass': 'M_act'})
        data = data.rename(columns={'Mini': 'M_ini'})
        data = data.rename(columns={'logL': 'logL/Lo'})

        # get unique ages and metallicities
        uage = sorted(data['Age'].unique())
        ufeh = sorted(data['FeH'].unique())

        # loop
        isochrones = {}
        for age in uage:
            for feh in ufeh:
                # get subset
                d = data[(data['Age'] == age) & (data['FeH'] == feh)]
                if len(d) == 0:
                    continue

                # get age and metallicity
                logging.info('Found Age=%.2fgyr, [M/H]=%.2f.', age/1e9, feh)

                # get meta data
                meta = {
                    'Age': age,
                    '[M/H]': feh,
                    'Type': 'CMD 2.9'
                }

                # create isochrone
                iso = Isochrone(data=d[['Teff', 'logg', 'logL/Lo', 'mbol', 'M_ini', 'M_act'] + filters], meta=meta)
                isochrones[(age, feh, 0)] = iso

        # finished
        return isochrones


class CMD33IsochroneFile(MultiIsochrone):
    """Isochrone file from PARSEC 3.3"""

    def _load_isochrones(self, filename: str) -> Dict[Tuple[float, float, float], Isochrone]:
        """Actually load isochrones from file.

        Args:
            filename: Name of file to load.

        Returns:
            Dictionary wuth isochrones.
        """

        # if we don't find an Av, we set it to 0
        av = 0

        # open file and find header and stuff
        with open(filename, 'r') as f:
            for line in f:
                # Av?
                if 'Av=' in line:
                    tmp = line[line.index('Av=') + 3:]
                    av = float(tmp[:tmp.index(',')])
                # find header
                if line.startswith('# Zini'):
                    names = line[2:].split()
                    break
            else:
                raise ValueError('Could not find header in isochrone file.')

        # load data
        data = pd.read_csv(filename, index_col=False, delim_whitespace=True, comment='#', names=names)

        # age and Teff
        data['logAge'] = 10. ** data['logAge']
        data['logTe'] = 10. ** data['logTe']

        # rename some columns
        data.rename(columns={'Mass': 'M_act', 'Mini': 'M_ini', 'logL': 'logL/Lo', 'logAge': 'Age',
                             'MH': 'FeH', 'logTe': 'Teff'},
                    inplace=True)

        # find filter columns and remove "mag" from names
        filters = [c[:-3] for c in data.columns if c.endswith('mag')]
        logging.info('Found filters: ' + ', '.join(filters))
        data.rename(columns={f + 'mag': f for f in filters}, inplace=True)

        # loop unique ages and metallicities
        isochrones = {}
        for age in sorted(data['Age'].unique()):
            for feh in sorted(data['FeH'].unique()):
                # get subset
                d = data[(data['Age'] == age) & (data['FeH'] == feh)]
                if len(d) == 0:
                    continue

                # get age and metallicity
                logging.info('Found Age=%.2fgyr, [M/H]=%.2f.', age/1e9, feh)

                # get meta data
                meta = {
                    'Age': age,
                    '[M/H]': feh,
                    'Av': av,
                    'Type': 'CMD 3.3'
                }

                # create isochrone
                iso = Isochrone(data=d[['Teff', 'logg', 'logL/Lo', 'M_ini', 'M_act'] + filters], meta=meta)
                isochrones[(age, feh, av)] = iso

        # finished
        return isochrones


__all__ = ['Isochrone', 'MultiIsochrone']



from typing import List

from spexxy.object import spexxyObject


class MainRoutine(spexxyObject):
    """MainRoutine is the base class for all main routines."""

    def __init__(self,  *args, **kwargs):
        """Initialize a new MainRoutine object"""
        spexxyObject.__init__(self, *args, **kwargs)

    def __call__(self):
        """Start the routine."""
        raise NotImplementedError


class FilesRoutine(MainRoutine):
    """FilesRoutine is the base class for all routines fitting files."""

    def __init__(self, *args, **kwargs):
        """Initialize a new MainRoutine object"""
        MainRoutine.__init__(self, *args, **kwargs)

    def parameters(self) -> List[str]:
        """Get list of parameters fitted by this routine.

        Returns:
            List of parameter names (including prefix) fitted by this routine.
        """
        return []

    def columns(self) -> List[str]:
        """Get list of columns returned by __call__.

        The returned list shoud include the list from parameters().

        Returns:
            List of columns returned by __call__.
        """

        # build list of columns from parameters
        columns = []
        for p in self.parameters():
            columns += [p.upper(), p.upper() + ' ERR']
        return columns

    def __call__(self, filename: str) -> List[float]:
        """Start the routine on the given file.

        Args:
            filename: Name of file to process.

        Returns:
            List of final values of parameters, ordered in the same way as the return value of parameters()
        """
        raise NotImplementedError


__all__ = ['MainRoutine', 'FilesRoutine']



import logging

from ..interpolator import Interpolator


def create_interpolator(definition: str, logger: logging.Logger = None) -> Interpolator:
    """Create an interpolator from a well defined <definition>, which is
    defined as:
        <definition>:<option>(,<option>)*

    Args:
        definition: Definition of grid.
        logger: If provided, log some status output.

    Returns:
        Instance of Interpolator object.
    """
    from .. import interpolator
    from .. import grid

    # first we split by ':'
    s = definition.split(':')
    if len(s) != 2:
        raise ValueError('Grid definition must be of form <type>:<options>.')

    # check type
    if s[0].lower() == 'ulyss':
        # create ulyss interpolator
        if logger:
            logger.info("Loading ulyss interpolator {0:s}.".format(s[1]))
        return interpolator.UlyssInterpolator(s[1])

    elif s[0].lower() == 'files':
        # further split is by ','
        t = s[1].split(',')

        # load grid
        if logger:
            logger.info('Loading phoenix grid {0:s}.'.format(t[0]))
        phxgrid = grid.FilesGrid(t[0])

        # second option would be for the spline interpolator
        if len(t) == 2:
            # if it is 'none', we do not load pre-calculated 2nd derivatives
            if t[1].lower() != 'none':
                if logger:
                    logger.info('Loading phoenix derivatives grid {0:s}.'.format(t[1]))
                derivs = grid.FilesGrid(t[1])
            else:
                if logger:
                    logger.warning('Using no pre-calculated derivatives.')
                derivs = None

            # return spline interpolator
            if logger:
                logger.info('Creating spline interpolator.')
            return interpolator.SplineInterpolator(phxgrid, derivs)

        else:
            # just return the linear interpolator
            if logger:
                logger.info('Creating linear interpolator.')
            return interpolator.LinearInterpolator(phxgrid)

    else:
        # invalid grid type
        if logger:
            logger.error("Invalid grid type {0:s}.".format(s[0]))
        raise ValueError("Invalid grid type {0:s}.".format(s[0]))



from .grid import GridComponent
from ..interpolator import Interpolator


class TelluricsComponent(GridComponent):
    """A Component for serving tellurics spectra from a given interpolator."""

    def __init__(self, interpolator: Interpolator, name: str = "TELLURICS", *args, **kwargs):
        """Initializes a new Grid component.

        Parameters
        ----------
        interpolator : Interpolator
            The interpolator to use for the component
        name : str
            Name of the component
        """
        GridComponent.__init__(self, interpolator, name, vac_to_air=True, *args, **kwargs)


__all__ = ['TelluricsComponent']



import os
import numpy as np
from astropy.io import fits
from spexxy.data.resultsfits import ResultsFITS
from spexxy.data.spectrum import SpectrumFitsHDU


class FitsSpectrum(object):
    """A class for handling both spectra and results in a FITS file."""

    def __init__(self, filename: str, mode: str = 'r'):
        """Open an existing or create a new FITS file.

        Args:
            filename: Name of file to open.
            mode: Open mode (w=write, r=read, rw=read/write).
        """

        # store it
        self._filename = filename
        self._mode = mode
        self._hdu_as_spectrum = {}

        # check mode
        if mode == 'r':
            # open fits file
            self._fits = fits.open(filename, memmap=False)

        elif mode == 'w':
            # create an empty HDU list
            self._fits = fits.HDUList()

        elif mode == 'rw':
            # open fits file in update mode
            self._fits = fits.open(filename, memmap=False, mode='update')

        else:
            raise ValueError('Invalid mode for opening FITS file.')

    def __enter__(self):
        """Returns just itself, when class is used in with statement."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """When used in with statement, leaving it closes the file."""
        self.close()

    def close(self):
        """Close the file."""

        # check mode
        if self._mode == 'r':
            # just close file
            self._fits.close()

        elif self._mode == 'w':
            # save file
            if os.path.exists(self._filename):
                os.remove(self._filename)
            self._fits.writeto(self._filename)
            self._fits.close()

        elif self._mode == 'rw':
            # flush and close
            self._fits.flush()
            self._fits.close()

    def save(self, filename: str = None):
        """Save spectrum to file.

        Args:
            filename: If filename is given, save to new file, otherwise use existing one (if exists).
        """

        # update HDUs
        for hdu, spec in self._hdu_as_spectrum.items():
            spec.hdu()

        # no filename given?
        filename = filename or self._filename
        if not filename:
            raise ValueError('No filename given for saving spectrum.')

        # save to new filename
        if os.path.exists(filename):
            os.remove(filename)
        self._fits.writeto(filename)

        # in any case, we change mode and filename
        self._filename = filename

    @property
    def filename(self):
        """Returns filename of spectrum, or None."""
        return self._filename

    def hdu_names(self):
        """Returns names of HDUs in FITS file."""
        return [hdu.header['EXTNAME'] if 'EXTNAME' in hdu.header else ''
                for hdu in self._fits]

    def hdu(self, hdu: int):
        """Returns a HDU, either by number or by name.

        Args:
            hdu: Number or name of HDU to return.

        Returns:
            Selected HDU.
        """

        return self._fits[hdu]

    def __len__(self):
        """Return number of HDUs in FITS file."""
        return len(self._fits)

    def __contains__(self, hdu):
        """Whether a HDU exists."""
        return hdu in self._fits

    def __getitem__(self, hdu):
        """Returns spectrum for given HDU.

        Args:
            param:  HDU to return spectrum for.

        Returns:
            A SpectrumFitsHDU object.
        """

        # get hdu
        fits_hdu = self.hdu(hdu)

        # stored already?
        if fits_hdu in self._hdu_as_spectrum:
            return self._hdu_as_spectrum[fits_hdu]

        # match data?
        if 'CRVAL1' not in fits_hdu.header and 'WAVE' not in fits_hdu.header:
            # copy, if possible
            if 'CRVAL1' in self._fits[0].header:
                fits_hdu.header['CRVAL1'] = self._fits[0].header['CRVAL1']
                fits_hdu.header['CDELT1'] = self._fits[0].header['CDELT1']
            elif 'WAVE' in self._fits[0].header:
                fits_hdu.header['WAVE'] = self._fits[0].header['WAVE']
            else:
                raise ValueError('No wavelength info found.')

        # create spectrum and return it
        self._hdu_as_spectrum[fits_hdu] = SpectrumFitsHDU(fits_hdu, hdu_list=self._fits, filename=self._filename)
        return self._hdu_as_spectrum[fits_hdu]

    def __setitem__(self, hdu_name, spec):
        """Sets a new spectrum for a HDU.

        Args:
            hdu_name: Name of HDU.

        Returns:
            Spectrum to set for that HDU.
        """

        # spec must be a SpectrumFitsHDU instance
        if not isinstance(spec, SpectrumFitsHDU):
            raise ValueError('Spectrum must be instance of SpectrumFitsHDU.')

        # check type
        if hdu_name == 0 and not spec.primary:
            raise ValueError('Cannot store non-primary HDU as 0.')
        if hdu_name != 0 and spec.primary:
            raise ValueError('Cannot store primary HDU with other name than 0.')

        # get hdu
        spec_hdu, wave_hdu = spec.hdu()

        # is it a primary HDU?
        if spec.primary:
            # yes, find and remove existing one
            prim = list(filter(lambda hdu: isinstance(hdu, fits.PrimaryHDU), self._fits))
            if len(prim) > 0:
                # remove from list of HDUs
                self._fits.remove(prim[0])
                # remove from mapping
                if prim[0] in self._hdu_as_spectrum:
                    del self._hdu_as_spectrum[prim[0]]

            # add it
            self._fits.insert(0, spec_hdu)
            self._hdu_as_spectrum[spec_hdu] = spec

        else:
            # extension HDU
            # name taken already?
            if hdu_name in self._fits:
                # get hdu of that name
                exist_hdu = self.hdu(hdu_name)

                # and the owner is not the given spectrum?
                if spec_hdu != exist_hdu:
                    # get index and delete HDU
                    del self._fits[hdu_name]
                    if exist_hdu in self._hdu_as_spectrum:
                        del self._hdu_as_spectrum[exist_hdu]

            # rename it
            spec_hdu.name = hdu_name

            # add to lists
            self._fits.append(spec_hdu)
            self._hdu_as_spectrum[spec_hdu] = spec

    def __delitem__(self, hdu_name):
        if hdu_name in self._fits:
            del self._fits[hdu_name]

    def append(self, hdu):
        self._fits.append(hdu)

    def wave_count(self):
        """Returns number of elements in spectrum."""
        spec = self.spectrum
        return len(spec)

    @property
    def spectrum(self):
        """Returns main spectrum, short vor FitsSpectrum[0]"""
        return self[0]

    @spectrum.setter
    def spectrum(self, spec):
        """Sets main spectrum, short vor FitsSpectrum[0]=spec"""
        self[0] = spec

    def results(self, namespace):
        """Returns a ResultsFITS object for the given namespace.

        Args:
            namespace: Namespace for the results.

        Returns:
            ResultsFITS object.
        """
        return ResultsFITS(self.hdu(0), namespace)

    @property
    def sigma(self):
        """Returns uncertainties for spectrum, or array of ones if they don't exist."""
        if 'SIGMA' in self.hdu_names():
            return self['SIGMA'].flux
        else:
            return np.ones((self.wave_count()))

    @property
    def good_pixels(self):
        """Returns mask of valid for spectrum, or array of ones if they don't exist."""
        if 'GOODPIXELS' in self.hdu_names():
            return self['GOODPIXELS'].flux.astype(bool)
        else:
            return np.ones((self.wave_count()), dtype=bool)

    @good_pixels.setter
    def good_pixels(self, arr: np.ndarray):
        """Sets mask of valid pixels in spectrum.

        Args:
            arr: New mask of valid pixels.
        """

        # create new Spectrum and set mask as flux
        spec = SpectrumFitsHDU(spec=self.spectrum, dtype=np.int16, primary=False)
        spec.flux = arr.astype(bool)

        # write into HDU
        self['GOODPIXELS'] = spec

    @property
    def residuals(self):
        """Returns residuals of spectrum fit or None if there are none."""
        return self['RESIDUALS'].flux if 'RESIDUALS' in self.hdu_names() else None

    @residuals.setter
    def residuals(self, arr: np.ndarray):
        """Sets residuals.

        Args:
            arr: New residuals for spectrum.
        """

        # create new Spectrum and set residuals as flux
        spec = SpectrumFitsHDU(spec=self.spectrum, primary=False)
        spec.flux = arr

        # write into HDU
        self['RESIDUALS'] = spec

    @property
    def mult_poly(self):
        """Returns multiplicative polynomial from spectrum fitting, or array of ones if they don't exist."""
        if 'MULTPOLY' in self.hdu_names():
            return self['MULTPOLY'].flux
        else:
            return np.ones((self.wave_count()))

    @mult_poly.setter
    def mult_poly(self, arr: np.ndarray):
        """Sets new multiplicative polynomial.

        Args:
            arr: New multiplicative polynomial for spectrum.
        """

        # create new Spectrum and set residuals as flux
        spec = SpectrumFitsHDU(spec=self.spectrum, primary=False)
        spec.flux = arr

        # write into HDU
        self['MULTPOLY'] = spec

    @property
    def best_fit(self):
        """Returns best fit of spectrum fit or None if there is none."""
        return self['BESTFIT'] if 'BESTFIT' in self.hdu_names() else None

    @best_fit.setter
    def best_fit(self, best: np.ndarray):
        """Sets new best fit.

        Args:
            best:  New best fit for spectrum.
        """

        # create new Spectrum and write it into HDU
        spec = SpectrumFitsHDU(spec=best, primary=False)
        self['BESTFIT'] = spec

    @property
    def tellurics(self):
        """Returns tellurics of spectrum fit or None if there is none."""
        return self['TELLURICS'] if 'TELLURICS' in self.hdu_names() else None

    @property
    def header(self):
        """Returns FITS header."""
        return self._fits[0].header

    @property
    def covar(self):
        """Returns covariance matrix."""
        return

    @covar.setter
    def covar(self, cov: np.ndarray):
        """Sets new covariance matrix.

        Args:
            cov: New covariance matrix for fit.
        """
        # name taken already?
        if 'COVAR' in self._fits:
            del self._fits['COVAR']

        # create HDU
        hdu = fits.ImageHDU(cov)
        hdu.name = 'COVAR'

        # add to list
        self._fits.append(hdu)


__all__ = ['FitsSpectrum']



import os

from .init import Init
from ..data import FitsSpectrum
from ..component import Component


class InitFromPath(Init):
    """Initializes a component from another file in a given directory.

    This class, when called, initializes the parameters of a given component to the values in the header of
    a file with the same name (usually written in a previous fit) in the given directory.
    """

    def __init__(self, path: str, *args, **kwargs):
        """Initializes a new Init object.

        Args:
            path: Path in which to look for the file to read the initial values from.
        """
        Init.__init__(self, *args, **kwargs)
        self._path = path

    def __call__(self, cmp: Component, filename: str):
        """Initializes parameters of the given component with values from another file.

        Args:
            cmp: Component to initialize.
            filename: Name of file (in given path) to read initial values from.
        """

        # get filename
        fn = os.path.join(self._path, filename)
        if not os.path.exists(fn):
            self.log.error('Could not find file %s for reading initials.', fn)
            return

        # open file
        self.log.info('Reading initial values for component "%s" from %s...', cmp.prefix, fn)
        with FitsSpectrum(fn) as fs:
            # get results for this component
            results = fs.results(cmp.prefix)

            # loop parameters of component
            for param in cmp.param_names:
                # is it in results?
                if param in results:
                    # get value
                    value = results[param][0]

                    # set it
                    self.log.info('Setting initial value for "%s" of component "%s" to %f...', param, cmp.prefix, value)
                    cmp[param] = value


__all__ = ['InitFromPath']



from ..object import spexxyObject
from ..component import Component


class Init(spexxyObject):
    """Init is the base class for all objects that can initialize values of a component."""

    def __init__(self, *args, **kwargs):
        """Initialize an Init object."""
        spexxyObject.__init__(self, *args, **kwargs)

    def __call__(self, cmp: Component, filename: str):
        """Initializes values for the given component.

        Parameters
        ----------
        cmp : Component
            Component to initialize.
        filename : str
            Name of file containing spectrum to create mask for.
        """
        raise NotImplementedError



from typing import Union

from .spectrumcomponent import SpectrumComponent
from ..data import Spectrum


class StarComponent(SpectrumComponent):
    def __init__(self, spec: Union[Spectrum, str], name: str = "STAR", *args, **kwargs):
        """Initializes a new Star component that just serves a single given spectrum.

        Parameters
        ----------
        spec : SpectrumComponent | str
            A Spectrum object or the filename of a spectrum to load
        name : str
            Name of new component
        """
        SpectrumComponent.__init__(self, name, *args, **kwargs)

        # load or copy spectrum?
        if isinstance(spec, Spectrum):
            self._spectrum = spec.copy(copy_flux=True)
        elif isinstance(spec, str):
            self._spectrum = Spectrum.load(spec)
        else:
            raise ValueError('Unknown type for spec.')

    def _model_func(self):
        """Get spectrum.

        Returns
        -------
        SpectrumComponent
            The spectrum that the component has been initialized with
        """
        return self._spectrum.copy()


__all__ = ['StarComponent']



import glob
import io
import os
import subprocess
from typing import Union, List
import pandas as pd
import numpy as np


def bulk_read_header(folder_or_files: Union[str, List], keys: Union[str, List[str]], pattern: str = '*.fits',
                     chunksize: int = 10000) -> pd.DataFrame:
    """
    A wrapper around the dfits/fitsort command-line tools to query a list of
    FITS-files for one or more header keywords and store their values in a
    pandas DataFrame.

    Args:
        folder_or_files: List of files or the folder to work in.
        keys: The header keyword(s) that should be queried.
        pattern: A common pattern by which the input FITS-files can be recognized.
        chunksize: Maximum number of files to process per system call.

    Returns:
        The values of the requested keywords in each FITS-file. Note that
        currently no conversion is performed on the columns, this has to be
        done separately. For example, to convert the values of column 'XXX'
        to numeric values, one can do:
            output['XXX'] = output['XXX'].apply(pd.to_numeric, errors='coerce')
    """

    # get list of files
    if isinstance(folder_or_files, str):
        files = [os.path.split(f)[1] for f in glob.glob(os.path.join(folder_or_files, pattern))]
        folder = folder_or_files
    elif isinstance(folder_or_files, list):
        files = folder_or_files
        folder = os.path.abspath(os.path.dirname(files[0]))
    else:
        raise ValueError('Unknown format.')

    # no chunksize given?
    if chunksize is None:
        chunksize = len(files)

    # prepare results
    results = []

    # loop chunks
    for start in np.arange(0, len(files), chunksize):
        # get subset
        sub = files[start:start + chunksize]

        # prepare dfits call
        arguments = ['dfits']
        arguments.extend(sub)
        dfits = subprocess.Popen(arguments, stdout=subprocess.PIPE, cwd=folder)

        # prepare fitsort call
        arguments = ['fitsort']
        if hasattr(keys, "__iter__") and not isinstance(keys, str):
            for key in keys:
                arguments.append(key)
        else:
            arguments.append(keys)
        fitsort = subprocess.Popen(arguments, stdin=dfits.stdout, stdout=subprocess.PIPE)

        # get output
        result = fitsort.communicate()[0]

        # convert output to pandas dataframe
        output = pd.read_table(io.StringIO(result.decode('utf-8')))
        for column in output.columns:
            if column[:7] == 'Unnamed':  # ignore columns named 'Unnamed ...'
                del output[column]
            elif column != column.strip():  # remove leading/trailing whitespaces in column names
                output.rename(columns={column: column.strip()}, inplace=True)

        # add to list
        results.append(output)

    # return concatenated results
    return pd.concat(results)



import yaml
import argparse
import logging
import sys
from collections import OrderedDict

from spexxy.application import Application
from spexxy.utils.log import setup_log


def ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict):
    """Work-around for loading the config YAML file into an OrderedDict.
    Taken from: https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
    Should not be required in Python 3.7+, where a dict is always sorted."""
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping)
    return yaml.load(stream, Loader=OrderedLoader)


def main():
    # init logging
    logging.getLogger().setLevel(logging.DEBUG)
    setup_log('spexxy.main', 'spexxy.log', mode='w')

    # init parser
    parser = argparse.ArgumentParser(description='spexxy spectrum fitting tool')
    parser.add_argument('config', help='spexxy configuration file', type=argparse.FileType('r'))
    parser.add_argument('filenames', help='filenames of spectra to fit', type=str, nargs='*')
    parser.add_argument('--mp', type=int, help='run fit in parallel in the given number of processes')
    parser.add_argument('-o', '--output', help='CSV to write results into', type=str, default='spexxy.csv')
    parser.add_argument('-r', '--resume', help='resume previous fit', action='store_true')

    # parse args
    args = parser.parse_args()

    # parse config, use ordered_load, if version<3.7
    if sys.version_info >= (3, 7):
        config = yaml.load(args.config, Loader=yaml.FullLoader)
    else:
        config = ordered_load(args.config)

    # create app and run it
    app = Application(config, filenames=args.filenames, ncpus=args.mp, output=args.output, resume=args.resume)
    app.run()


if __name__ == '__main__':
    main()



import os
import logging
from astropy.io import fits
import numpy as np

from .weight import Weight
from ..data import Spectrum


class WeightFromSNR(Weight):
    """Reads the S/N ratio from the given file and creates weights from it as 1/SQRT(FLUX/SNR).

    This class, when called, loads the SNR from the given file and returns weights from it as 1/SQRT(FLUX/SNR).
    """

    def __init__(self, keyword: str = 'HIERARCH SPECTRUM SNRATIO', *args, **kwargs):
        """Initializes a new weight.

        Args:
            keyword: FITS header keyword containing S/N.
        """
        Weight.__init__(self, *args, **kwargs)
        self._keyword = keyword

    def __call__(self, spectrum: Spectrum, filename: str) -> np.ndarray:
        """Creates a new weight for a spectrum from its S/N.

        Args:
            spectrum: Spectrum to create weight for.
            filename: Name of file containing spectrum to create weight for.

        Returns:
            Array containing weight for given spectrum.
        """

        self.log.info('Creating weights from S/N...')

        # get S/N
        snr = fits.getval(filename, self._keyword)

        # calculate weights
        return 1. / np.sqrt(spectrum.flux / snr)


__all__ = ['WeightFromSNR']



import numpy as np


def clip_sigma(data: np.ndarray, kappa_low: float = 5, kappa_high: float = 5, maxiter: int = 99) -> np.ndarray:
    """Kappa-sigma clipping on numpy array.

    Args:
        data: Input array.
        kappa_low: Lower kappa.
        kappa_high: Upper kappa.
        maxiter: Maximum number of iterations.

    Returns:
        Clipped array.
    """

    # do iterations
    for i in range(maxiter):
        # calculate current median and sigma
        median = np.nanmedian(data)
        sigma = np.nanstd(data)

        # create mask
        mask = (data < (median - kappa_low * sigma)) | (data > (median + kappa_high * sigma))

        # no new pixels masked?
        if np.sum(mask) == 0:
            break

        # apply mask
        data = data[~mask]

    # return clipped array
    return data



from pathlib import Path
from single_source import get_version


__version__ = get_version("pyobs-core", Path(__file__).parent.parent)


def version() -> str:
    return "0.0.0" if __version__ is None else __version__


__all__ = ["version", "__version__"]



from astropy.io import fits
import numpy as np
import os
from typing import List, Tuple

from . import Interpolator
from ..grid import GridAxis, Grid
from ..data import Spectrum


class TelluricsInterpolator(Interpolator):
    """A tellurics interpolator"""

    def __init__(self, path: str, molecules: list, *args, **kwargs):
        """Initializes a new tellurics interpolator.

        Args:
            path: Path to files containing telluric grids
            molecules: Name of telluric grid files without .fits extension.

        Examples:
            Initialize the interpolator in the given path ($SPEXXYPATH/Tellurics) with H2O and O2 as the fitted
            molecules, provided by two different files (TelluricsH2O_R10000.fits, TelluricsO2_R10000.fits) in
            that directory:
                class: spexxy.interpolator.TelluricsInterpolator
                path: $SPEXXYPATH/Tellurics
                molecules: [TelluricsH2O_R10000, TelluricsO2_R10000]
        """

        Interpolator.__init__(self, *args, **kwargs)

        # load grids
        self._grids = []
        self._axes = []

        # load grids/axes
        self.log.info('Initializing tellurics interpolator...')
        for i, molec in enumerate(molecules):
            # get filename
            filename = os.path.join(os.path.expandvars(path), molec + '.fits')

            # load coefficients
            self.log.info("Loading tellurics grid from " + filename)
            self._grids += [fits.getdata(filename)]

            # get some header infos
            hdr = fits.getheader(filename)

            # add axis
            self._axes.append(
                GridAxis(name=hdr['MOLECULE'], min=hdr['VALIDMIN'], max=hdr['VALIDMAX'], initial=0.)
            )

            # get info from first grid
            if i == 0:
                self._wStart = hdr["CRVAL1"]
                self._wStep = hdr["CDELT1"]
                self._waveMode = Spectrum.Mode.LAMBDA
                if hdr["CTYPE1"] == "AWAV-LOG":
                    self._waveMode = Spectrum.Mode.LOGLAMBDA

    def axes(self) -> List[GridAxis]:
        """Returns information about the axes.

        Returns:
            List of GridAxis objects describing the grid's axes.
        """
        return self._axes

    def __call__(self, params: Tuple) -> Spectrum:
        """Interpolates at the given parameter set.

        Args:
            params: Parameter set to interpolate at.

        Returns:
            Interpolated spectrum at given position.
        """

        # loop params
        flux = None
        for i in range(len(params)):
            f = self._interpol(self._grids[i], params[i])
            if flux is None:
                flux = f
            else:
                flux *= f

        # return spec
        return Spectrum(flux=flux, wave_start=self._wStart, wave_step=self._wStep, wave_mode=self._waveMode)

    def _interpol(self, grid: Grid, value: Tuple) -> np.ndarray:
        """Do the actual interpolation in the given grid at the given position.

        Args:
            grid: Grid to interpolate in.
            params: Parameter to interpolate at.

        Returns:
            Interpolated tellurics at given position.
        """

        # matrix
        mat = np.empty((grid.shape[0]))
        for i in range(grid.shape[0]):
            if i == 0:
                mat[0] = 1.
            else:
                mat[i] = mat[i - 1] * value

        # evaluate
        return np.dot(mat, grid)


__all__ = ['TelluricsInterpolator']



import glob
import logging
import multiprocessing
import os
import time
import gc
import pandas as pd
from spexxy.interpolator import Interpolator

from .main import MainRoutine, FilesRoutine
from .object import create_object
from .utils.log import setup_log, shutdown_log


class Application(object):
    def __init__(self, config, filenames=None, ncpus=None, output=None, resume=False):
        # store it
        self._config = config
        self._filenames = filenames
        self._ncpus = ncpus
        self._output = output
        self._resume = resume

    def run(self):
        """Run application."""

        # init objects
        log = logging.getLogger('spexxy.main')
        log.info('Creating all objects...')
        objects = self._create_objects(log=log)
        log.info('Finished creating objects.')

        # initialize main routine
        log.info('Initializing main routine...')
        main = create_object(self._config['main'], objects=objects, log=log)

        # what type is main?
        if isinstance(main, FilesRoutine):
            # run on files
            self._run_on_files(log, main)
        elif isinstance(main, MainRoutine):
            # just run it
            main()

    def _run_on_files(self, log: logging.Logger, main: FilesRoutine):
        """Run the given FilesRoutine

        Args:
            log: Logger to use
            main: Routine to run
        """

        # filenames?
        if self._filenames is None:
            log.info('Nothing to do, going to bed...')

        # expand list of spectra
        filenames = []
        for f in self._filenames:
            if '*' in f or '?' in f:
                filenames.extend(glob.glob(f))
            else:
                filenames.append(f)

        # get columns
        columns = main.columns()

        # output csv?
        if self._output is not None:
            if self._resume and os.path.exists(self._output):
                # loading pre-existing data
                log.info('Loading results from existing output file...')
                data = pd.read_csv(self._output, index_col=False)

                # do columns match?
                if columns != list(data.columns.values)[1:]:
                    log.error('Columns in existing output file do not match request, please delete it.')
                    raise RuntimeError

                # filter files
                log.info('Filtering finished files...')
                filenames = list(filter(lambda filename: filename not in data['Filename'].values, filenames))

            else:
                # write new header
                log.info('Writing new output file...')
                with open(self._output, 'w') as f:
                    f.write('Filename,' + ','.join(columns) + '\n')

        # sort and count files
        self._filenames = sorted(filenames)
        self._total = len(self._filenames)
        log.info('Found a total of %d files to process.', len(self._filenames))

        # anything to do?
        if self._total == 0:
            log.info('Nothing to do, going to bed...')
            return

        # init
        pool = None
        if self._ncpus is None:
            # no, run sequentially
            log.info("Running analysis sequentially.")

        else:
            # number of cpus
            nprocs = min(self._ncpus, self._total)

            # yes, create pool of workers
            log.info("Starting analysis in parallel on %d CPUs..." % nprocs)
            pool = multiprocessing.Pool(nprocs)

        # loop
        for i, filename in enumerate(self._filenames, 1):
            # parallel?
            if self._ncpus is None:
                self._run_single(i, filename)
            else:
                pool.apply_async(self._run_single, (i, filename), error_callback=self.error)

        # join pool
        if pool is not None:
            pool.close()
            pool.join()

        # finished
        log.info('Finished.')

    def error(self, exception):
        logging.getLogger('spexxy.main').error('Something went wrong.', exc_info=exception)

    def _run_single(self, idx, filename):
        # start time
        start_time = time.time()

        # main logger
        main_log = logging.getLogger('spexxy.main')

        # log
        main_log.info('(%i/%i) Starting on file %s...', idx, self._total, filename)

        # file exists?
        if not os.path.exists(filename):
            main_log.error('(%i/%i) File does not exist.', idx, self._total, filename)
            return

        # init file logger, show stdout output only if we're running on a single cpu
        log = setup_log('spexxy.fit', filename.replace('.fits', '.log'), stream=(self._ncpus is None))

        # create objects
        log.info('Creating all objects...')
        objects = self._create_objects(log=log)
        log.info('Finished creating objects.')

        # initialize main routine
        log.info('Initializing main routine...')
        main: FilesRoutine = create_object(self._config['main'], objects=objects, log=log)

        # init components
        log.info('Setting initial values...')
        if 'components' not in objects or objects['components'] is None:
            objects['components'] = {}
        for name, cmp in objects['components'].items():
            cmp.init(filename)

        # start fit
        results = None
        try:
            # do the fit
            log.info('Starting fit...')
            results = main(filename)
        except:
            log.exception('Exception during execution of fit.')

        # write result
        if self._output is not None and results is not None:
            with open(self._output, 'a') as f:
                # write filename
                f.write('%s' % filename)

                # write results
                if len(results) > 0:
                    f.write(',' + ','.join(['' if r is None else str(r) for r in results]))

                # write line break
                f.write('\n')

        # clear cache in all interpolators
        if 'interpolators' in objects:
            log.info('Clearing all interpolator caches...')
            cleared_caches = 0
            for o in objects['interpolators'].values():
                if isinstance(o, Interpolator):
                    o.clear_cache()
                    cleared_caches += 1
            log.info('Cleared %d cache(s).', cleared_caches)

        # clean up
        log.info('Cleaning up objects...')
        for group, elements in objects.items():
            elements.clear()
        main = None
        gc.collect()

        # shutdown logger
        main_log.info('(%i/%i) Finished file %s...', idx, self._total, filename)
        log.info('Finished fit in %.2f seconds.', time.time() - start_time)
        shutdown_log('spexxy.fit')

    def _create_objects(self, log=None):
        # create objects
        objects = {}
        for group, value in self._config.items():
            # don't create "main"
            if group != 'main':
                # all other groups are nested
                objects[group] = {}
                for name, config in value.items():
                    objects[group][name] = create_object(config, objects=objects, name=name, log=log)

        # finished
        return objects



import glob
import logging
import argparse
import os
from typing import List

from spexxy.data import FitsSpectrum, SpectrumFitsHDU

log = logging.getLogger(__name__)


def add_parser(subparsers):
    # init parser
    parser = subparsers.add_parser('extract', help='Extract wavelength range from spectrum',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('start', type=float, help='Start of wavelength range')
    parser.add_argument('end', type=float, help='End of wavelength range')
    parser.add_argument('input', type=str, help='Input spectrum', nargs='+')
    parser.add_argument('-p', '--prefix', type=str, help='Output file prefix', default='extracted_')

    # argparse wrapper for create_grid
    parser.set_defaults(func=lambda args: extract_spectra(**vars(args)))


def extract_spectra(input: List[str], prefix: str, start: float, end: float, **kwargs):
    # get list of filenames
    files = []
    for filename in input:
        if '*' in input or '?' in input:
            files.extend(glob.glob(filename))
        else:
            files.append(filename)

    # make unique
    files = sorted(list(set(files)))

    # loop files
    for i, filename in enumerate(files, 1):
        print('(%d/%d) %s' % (i, len(files), filename))
        extract_spectrum(filename, prefix + os.path.basename(filename), start, end)


def extract_spectrum(input: str, output: str, start: float, end: float):
    """Extracts the given wavelength range from input and store it as output.

    Args:
        input: Input filename
        output: Output filename
        start: Wavelength start
        end: Wavelength end
    """

    # load spectrum
    with FitsSpectrum(input) as fs_in:
        # open spectrum to write
        with FitsSpectrum(output, 'w') as fs_out:
            # copy extracted spectrum
            fs_out.spectrum = fs_in.spectrum.extract(start, end)

            # loop all extensions:
            for ext in fs_in.hdu_names():
                # skip no name
                if ext.strip() == '':
                    continue

                # try to get hdu
                try:
                    # get hdu
                    hdu = fs_in[ext]

                    # need to create new hdu with primary=False, since that gets lost on extract
                    fs_out[ext] = SpectrumFitsHDU(spec=hdu.extract(start, end), primary=False)

                except ValueError:
                    pass


import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d


def add_parser(subparsers):
    # init parser
    parser = subparsers.add_parser('interpolate', help='Interpolate isochrone')
    parser.set_defaults(func=run)

    # parameters
    parser.add_argument('input', help='Input file containing the isochrone', type=str)
    parser.add_argument('output', help='Output file', type=str)
    parser.add_argument('--count', help='Number of points for new isochrone', type=int, default=500)
    parser.add_argument('--column', help='Column used for interpolation', type=str, default='M_ini')
    parser.add_argument('--space', help='Space for distance calculations', type=str, nargs=2, default=['Teff', 'logg'])
    parser.add_argument('-p', '--plot', help='Plot result', action='store_true')


def _norm_space(s):
    s -= np.min(s)
    return s / np.max(s)


def run(args):
    # load data
    isochrone = pd.read_csv(args.input, index_col=False, comment='#')

    # get X column for interpolator
    x = isochrone[args.column]
    x_min = np.min(x)
    x_max = np.max(x)

    # normalize space to 0..1
    space_x = _norm_space(isochrone[args.space[0]].values.copy())
    space_y = _norm_space(isochrone[args.space[1]].values.copy())

    # integrate space along isochrone
    logging.info('Creating new grid with %d points in %s between %.3f and %.3f', args.count, args.column, x_min, x_max)
    dist = [0.]
    for i in range(len(space_x) - 1):
        dist += [dist[-1] + np.sqrt((space_x[i+1]-space_x[i])**2. + (space_y[i+1]-space_y[i])**2.)]

    # create interpolator of x on dist
    ip = interp1d(dist, x)

    # interpolate x with constant sampling, which will be our new grid
    grid = ip(np.linspace(0., dist[-1], args.count))

    # get new grid
    data = {args.column: grid}

    # interpolate
    for col in isochrone.columns:
        # skip column from command line
        if col == args.column:
            continue
        logging.info('Interpolating column "%s"...', col)

        # create interpolator
        ip = interp1d(x, isochrone[col].values)

        # do interpolation
        data[col] = ip(grid)

    # create new dataframe
    output = pd.DataFrame(data)

    # write output
    output[isochrone.columns].to_csv(args.output, index=False)

    # plot
    if args.plot:
        plt.scatter(isochrone['Teff'], isochrone['logg'], s=10.0, label='Input isochrone')
        plt.scatter(output['Teff'], output['logg'], s=0.5, label='Output isochrone')
        plt.legend()
        plt.grid()
        plt.gca().invert_xaxis()
        plt.gca().invert_yaxis()
        plt.xlabel(args.space[0])
        plt.ylabel(args.space[1])
        plt.show()



import os
import logging
from astropy.io import fits
import numpy as np
from typing import List, Tuple

from .weight import Weight
from ..data import Spectrum


class WeightRanges(Weight):
    """Creates a weights array from given ranges.

    This class, when called, creates a weights array from the given wavelength ranges.
    """

    def __init__(self, ranges: List[Tuple[float, float, float]] = None, initial: float = 1., *args, **kwargs):
        """Initializes a new weight.

        Args:
            ranges: List of tuples of (wave start, wave end, weight).
            initial: Initial value for whole array.
        """
        Weight.__init__(self, *args, **kwargs)
        self._ranges = ranges if ranges is not None else []
        self._initial = initial

    def __call__(self, spectrum: Spectrum, filename: str) -> np.ndarray:
        """Creates a new weight for a spectrum from the ranges given in the configuration.

        Args:
            spectrum: Spectrum to create weight for.
            filename: Name of file containing spectrum to create weight for.

        Returns:
            Array containing weight for given spectrum.
        """

        # init array
        self.log.info('Creating weights for spectrum from given ranges...')
        weights = np.zeros((len(spectrum))) + self._initial

        # loop ranges
        for start, end, weight in self._ranges:
            # get pixels in range
            w = (spectrum.wave >= start) & (spectrum.wave <= end)
            # set weight
            weights[w] = weight

        # finished
        return weights


__all__ = ['WeightRanges']



from typing import Union
from astropy.io import fits
import numpy as np

from .mask import Mask
from ..data import Spectrum


class MaskRanges(Mask):
    """Masks ranges in spectra.

    This class, when called, creates a mask on the given wavelength ranges
    """

    def __init__(self, ranges: list, vrad: Union[str, float] = None,
                 component: str = None, vrad_parameter: str = None,
                 *args, **kwargs):
        """Initializes a new mask.

        Args:
            ranges: List of tuples defining (start, end) of wavelength ranges to mask.
            vrad: Radial velocity to shift by, either a number or the name of a FITS header entry, in which
                case "-<name>" negates the value.
            component: Name of component to read the radial velocity from.
            vrad_parameter: Name of parameter in given component to use as radial velocity.
        """
        Mask.__init__(self, *args, **kwargs)
        self._ranges = ranges
        self._vrad = vrad
        self._component = component
        self._vrad_parameter = vrad_parameter

    @staticmethod
    def _redshift(wave: float, vrad: float, mode: Spectrum.Mode = Spectrum.Mode.LAMBDA) -> float:
        """Redshifts a given wavelengths by the given radial velocity.

        Args:
            wave: Wavelength to redshift.
            vrad: Radial velocity to shift by.
            mode: Reference frame of wavelength (LAMBDA or LOG).

        Returns:
            Redhifted wavelength.
        """

        # calculate redshift depending on mode
        if mode == Spectrum.Mode.LAMBDA:
            return wave * (1. + vrad / 299792.458)
        else:
            return wave + np.log(1. + vrad / 299792.458)

    def __call__(self, spectrum: Spectrum, filename: str = None) -> np.ndarray:
        """Creates a new dynamic mask for a spectrum in the wavelength ranges given in the configuration and
        shifted by the current radial velocity of the given component.

        Args:
            spectrum: Spectrum to create mask for.
            filename: Name of file containing spectrum to create mask for.

        Returns:
            Boolean array containing good pixel mask for given spectrum.
        """

        # by default, there is no vrad shift
        vrad = 0.

        # get vrad from component
        if self._component is not None and self._vrad_parameter is not None:
            vrad = self.objects['components'][self._component][self._vrad_parameter]

        # additional vrad?
        if self._vrad is not None:
            # str or float?
            if isinstance(self._vrad, float):
                # absolute value
                self.log.info('Shifting all dynamic masks by constant %.2f km/s.', self._vrad)
                vrad += self._vrad
            elif isinstance(self._vrad, str):
                # FITS header
                mask_sign = 1.
                mask_shift = self._vrad
                # negative?
                if self._vrad[0] == '-':
                    mask_sign = -1
                    mask_shift = self._vrad[1:]
                # get fits header
                hdr = fits.getheader(filename)
                try:
                    # get shift and add it
                    val = mask_sign * hdr[mask_shift]
                    self.log.info('Shifting all masks by %.3f km/s.', val)
                    vrad += val
                except KeyError:
                    self.log.error('Could not find FITS header "%s" in file.')

        # init mask
        mask = np.ones((len(spectrum.wave)), dtype=bool)

        # loop ranges
        for rng in self._ranges:
            # need to swap?
            if rng[0] > rng[1]:
                rng = (rng[1], rng[0])

            # shift start/end
            rng = (MaskRanges._redshift(rng[0], vrad), MaskRanges._redshift(rng[1], vrad))

            # create (inverted) mask
            m = (spectrum.wave >= rng[0]) & (spectrum.wave <= rng[1])

            # apply it
            mask &= ~m

            # log it
            self.log.info('Masked range %.2f-%.2f AA (%d pixels).', rng[0], rng[1], np.sum(m))

        # finished
        return mask


__all__ = ['MaskRanges']



import numpy as np

from ..object import spexxyObject
from ..data import Spectrum


class Weight(spexxyObject):
    """Weight is the base class for all objects that can create a weight array for spectra."""

    def __init__(self, *args, **kwargs):
        """Initialize a new weight."""
        spexxyObject.__init__(self, *args, **kwargs)

    def __call__(self, spectrum: Spectrum, filename: str) -> np.ndarray:
        """Creates a new weight for a spectrum.

        Args:
            spectrum: Spectrum to create weight for.
            filename: Name of file containing spectrum to create weight for.

        Returns:
            Array containing weight for given spectrum.
        """
        raise NotImplementedError


__all__ = ['Weight']



from typing import List, Tuple, Dict

from . import Interpolator
from ..grid import GridAxis
from ..data import Spectrum


class SubInterpolator(Interpolator):
    """An interpolator that only gives access to a part of the parent interpolator."""

    def __init__(self, parent: Interpolator, fixed: Dict[str, float], *args, **kwargs):
        """Initializes a new sub interpolator.

        Args:
            parent: Parent interpolator.
            fixed: Dictionary with fixed axis_name->value pairs.
        """
        Interpolator.__init__(self, *args, **kwargs)
        self.log.info('Initializing sub interpolator...')

        # store
        self._parent = parent
        self._fixed = fixed

        # get axes
        self._axes: List[GridAxis] = []
        for ax in parent.axes():
            if ax.name in fixed.keys():
                self.log.info('Fixing axis %s to value %f.', ax.name, fixed[ax.name])
            else:
                self._axes.append(ax)

    def axes(self) -> List[GridAxis]:
        """Returns information about the axes.

        Returns:
            List of GridAxis objects describing the grid's axes
        """
        return self._axes

    def __call__(self, params: Tuple) -> Spectrum:
        """Interpolates at the given parameter set

        Args:
            params: Parameter set to interpolate at

        Returns:
            Interpolated spectrum at given position
        """

        # convert given params to list
        given_params = list(params)

        # loop all parameters
        full_params = []
        for ax in self._parent.axes():
            if ax.name in self._fixed.keys():
                # if axis is fixed, use fixed value
                full_params.append(self._fixed[ax.name])
            else:
                # otherwise pull next value from given_params
                full_params.append(given_params.pop(0))

        # call parent interpolator
        return self._parent(tuple(full_params))


__all__ = ['SubInterpolator']



from abc import abstractmethod
from typing import List, Any, Tuple, Union, Type, Protocol, Optional
import pandas as pd
import os

from .grid import Grid, GridAxis
from ..data import Spectrum
from ..object import get_class_from_string


class SpecLoader(Protocol):
    @staticmethod
    @abstractmethod
    def load(filename: str) -> Spectrum:
        raise NotImplementedError


class FilesGrid(Grid):
    """Grid working on files with a CSV based database."""

    def __init__(self, filename: str, norm_to_mean: bool = False, spec_loader: Optional[Union[str, SpecLoader]] = None,
                *args, **kwargs):
        """Constructs a new Grid.

        Args:
            filename: Filename of CSV file.
            norm_to_mean: Normalize spectra to their mean.
            spec_loader: Class to load spectrum, must provide load(str)->Spectrum method
        """

        # store
        self._norm_to_mean = norm_to_mean
        if spec_loader is None:
            self._spec_loader = Spectrum
        else:
            self._spec_loader = get_class_from_string(spec_loader) if isinstance(spec_loader, str) else spec_loader

        # expand filename
        filename = os.path.expandvars(filename)

        # get grid's root path
        self._root = os.path.dirname(filename)

        # load CSV
        self._data = pd.read_csv(filename, index_col=False)

        # get all parameters, by removing 'Filename' from list of columns
        parameters = list(self._data.columns)
        parameters.remove('Filename')

        # we assume that all parameters are floats, so treat them as such
        for name in parameters:
            self._data[name] = self._data[name].apply(lambda x: float(x))

        # create axes
        axes = [GridAxis(name=name, values=sorted(self._data[name].unique())) for name in parameters]

        # init grid
        Grid.__init__(self, axes, *args, **kwargs)

        # set index on dataframe
        self._data.set_index(parameters, inplace=True)

    def all(self) -> List[Tuple]:
        """Return all possible parameter combinations.

        Returns:
            All possible parameter combinations.
        """
        return self._data.index.values

    def __contains__(self, params: Tuple) -> bool:
        """Checks, whether the grid contains a given parameter set.

        Args:
            params: Parameter set to check.

        Returns:
            Whether or not the given parameter set exists in the grid.
        """
        return tuple(params) in self._data.index

    def filename(self, params: Tuple, absolute: bool = True) -> str:
        """Returns filename for given parameter set.

        Args:
            params: Parameter set to catch value for.
            absolute: If True, return full absolute path, otherwise relative path within grid.

        Returns:
            Filename.
        """

        if absolute:
            return os.path.join(self._root, self._data.loc[tuple(params)].Filename)
        else:
            return self._data.loc[tuple(params)].Filename

    def __call__(self, params: Tuple) -> Any:
        """Fetches the value for the given parameter set.

        Args:
            params: Parameter set to catch value for.

        Returns:
            Grid value at given position.
        """

        # get filename
        tmp = self._data.loc[tuple(params)]
        if isinstance(tmp, pd.core.frame.DataFrame):
            if len(tmp) > 0:
                self.log.warning('More than one matching spectrum found, taking first.')
            filename = tmp.iloc[0].Filename
        else:
            filename = tmp.Filename

        # load Spectrum
        spec = self._spec_loader.load(os.path.join(self._root, filename))

        # normalize?
        if self._norm_to_mean:
            spec.norm_to_mean()

        # return it
        return spec


__all__ = ['FilesGrid']



from typing import List, Tuple

import scipy.special
import scipy.linalg
import scipy.interpolate
import numpy as np
from numpy import polynomial


class Continuum(object):
    """Base class for all continuum fitting classes."""

    def __init__(self, poly: str = "polynomial", poly_degree: int = 7, spline_type: str = 'cubic'):
        """Initialize a new Continuum

        Args:
            poly: Type of polynomial ('polynomial', 'chebyshev', or 'legendre') or 'spline'.
            poly_degree: Degree of polynomial.
            spline_type: Type of spline ('quadratic', 'cubic')
        """

        # polynomial function or spline?
        if poly == 'spline':
            self._poly_func = None
        else:
            self._poly_func = {
                'polynomial': polynomial.Polynomial,
                'chebyshev': polynomial.Chebyshev,
                'legendre': polynomial.Legendre
            }[poly]

        # store
        self._poly_degree = poly_degree
        self._spline_type = spline_type

    def _fit_poly(self, xin: np.ndarray, yin: np.ndarray, xout: np.ndarray = None) -> np.ndarray:
        """Fit a polynomial to the data and return it.

        Args:
            xin: Input x array for continuum to fit.
            yin: Input y array for continuum to fit.
            xout: Output x array to evaluate polynomial on, if None is given, xin is used.

        Returns:
            Array containing calculated polynomial.

        """

        # no xout?
        if xout is None:
            xout = xin

        # do actual fitting
        if self._poly_func is not None:
            # fit with a polynomial
            result = self._poly_func.fit(xin, yin, deg=self._poly_degree)

            # evaluate polynomial at xout
            return result(xout)

        else:
            # fit with a spline
            ip = scipy.interpolate.interp1d(xin, yin, kind='cubic', bounds_error=False, fill_value='extrapolate')

            # evaluate spline at xout
            return ip(xout)

    def __call__(self, x: np.ndarray, y: np.ndarray, valid: np.ndarray = None) -> np.ndarray:
        """Calculate the continuum.

        Args:
            x: x array for continuum to fit.
            y: y array for continuum to fit.
            valid: Valid pixels mask.

        Returns:
            Array containing calculated continuum.
        """

        # no valid array?
        if valid is None:
            valid = np.ones((x.shape), dtype=bool)

        # fit polynomial
        return self._fit_poly(x[valid], y[valid], x)

    def remove(self, x: np.ndarray, y: np.ndarray, valid: np.ndarray = None) -> np.ndarray:
        """Remove continuum from a given array.

        Args:
            x: x array for continuum to fit.
            y: y array for continuum to fit.
            valid: Valid pixels mask.

        Returns:
            Array containing continuum corrected y array.
        """

        # calculate continuum
        cont = self(x, y, valid)

        # do we have negative values in the continuum?
        min_cont = np.min(cont)
        if min_cont < 1:
            # correct y by shifting it into positive range first
            return (y + min_cont + 1) / (cont + min_cont + 1)
        else:
            # otherwise remove continuum directly
            return y / cont


class MaximumBin(Continuum):
    """Derive continuum as spline to all points that are within a given fraction of largest points in a given number
    of bins."""

    def __init__(self, frac: float = 0.15, sbin: int = 100, *args, **kwargs):
        """Initialize a new MaximumBin continuum.

        Args:
            frac: Fraction of largest points in each bin to use for fitting the continuum.
            sbin: Number of bins.
        """
        Continuum.__init__(self, *args, **kwargs)

        # remember values
        self.frac = frac
        self.sbin = sbin

    def __call__(self, x: np.ndarray, y: np.ndarray, valid: np.ndarray = None) -> np.ndarray:
        """Calculate the continuum.

        Args:
            x: x array for continuum to fit.
            y: y array for continuum to fit.
            valid: Valid pixels mask.

        Returns:
            Array containing calculated continuum.
        """

        # no valid array?
        if valid is None:
            valid = np.ones((x.shape), dtype=bool)

        # find continuum for every of the sbin bins
        bins_x = np.zeros((self.sbin))
        bins_y = np.zeros((self.sbin))
        w1 = 0
        w2 = float(len(x)) / self.sbin
        for i in range(self.sbin):
            # get range
            bindata = y[int(w1):int(w2)].copy()

            # sort it
            bindata.sort()

            # calculate median
            bins_y[i] = np.median(bindata[int(-self.frac * len(bindata)):-1])

            # set wavelength
            bins_x[i] = np.mean(x[int(w1):int(w2)])

            # reset ranges
            w1 = w2
            w2 += float(len(x)) / self.sbin

            # check for last bin
            if i == self.sbin - 1:
                w2 = len(x)

        # fit polynomial
        return self._fit_poly(bins_x, bins_y, x)


class SigmaClipping(Continuum):
    """Derives continuum from kappa-sigma clipping with polynomial."""

    def __init__(self, kappa_low: float = 3, kappa_high: float = 3, maxiter: int = 20, *args, **kwargs):
        """Initialize a new SigmaClipping continuum

        Args:
            kappa_low: Lower kappa.
            kappa_high: Upper kappa.
            maxiter: Maximum number of iterations.
        """
        Continuum.__init__(self, *args, **kwargs)

        # store values
        self.kappa_low = kappa_low
        self.kappa_high = kappa_high
        self.maxiter = maxiter

    def __call__(self, x: np.ndarray, y: np.ndarray, valid: np.ndarray = None) -> np.ndarray:
        """Calculate the continuum.

        Args:
            x: x array for continuum to fit.
            y: y array for continuum to fit.
            valid: Valid pixels mask.

        Returns:
            Array containing calculated continuum.
        """

        # no valid array?
        if valid is None:
            valid = np.ones((x.shape), dtype=bool)

        # copy x and y
        xx = x[valid].copy()
        yy = y[valid].copy()

        # fit
        poly = self._fit_poly(xx, yy)

        # iterate
        i = 0
        while i < self.maxiter:
            # calc residuals and do sigma clipping
            residuals = yy - poly

            # calculate current median and sigma
            median = np.nanmedian(residuals)
            sigma = np.nanstd(residuals)

            # create mask
            mask = (residuals < (median - self.kappa_low * sigma)) | (residuals > (median + self.kappa_high * sigma))

            # no new pixels masked?
            if np.sum(mask) == 0:
                break

            # apply mask
            xx = xx[~mask]
            yy = yy[~mask]

            # fit new poly
            poly = self._fit_poly(xx, yy)

        # return continuum
        return self._fit_poly(xx, yy, xout=x)


class Regions(Continuum):
    """Derives continuum from given regions."""

    def __init__(self, regions: List[Tuple[float, float]], average: bool = True, *args, **kwargs):
        """Initializes new Regions continuum

        Args:
            regions: List of tuples with start end end x values for regions.
            average: Use average wave/flux in regions instead of all pixels.
        """

        Continuum.__init__(self, *args, **kwargs)
        self.regions = regions
        self.average = average

    def __call__(self, x: np.ndarray, y: np.ndarray, valid: np.ndarray = None) -> np.ndarray:
        """Calculate the continuum.

        Args:
            x: x array for continuum to fit.
            y: y array for continuum to fit.
            valid: Valid pixels mask.

        Returns:
            Array containing calculated continuum.
        """

        # no valid array?
        if valid is None:
            valid = np.ones(x.shape, dtype=bool)

        # evaluate regions
        xx = []
        yy = []
        for r in self.regions:
            # create mask
            mask = valid & (x >= r[0]) & (x <= r[1])

            # add (mean) of points within mask
            if self.average:
                xx.append(np.mean(x[mask]))
                yy.append(np.mean(y[mask]))
            else:
                xx.extend(x[mask])
                yy.extend(y[mask])

        # fit continuum
        return self._fit_poly(xx, yy, xout=x)


class PeakToPeak(Continuum):
    """Derives continuum from a ptp optimization.

    from "Spectroscopy of Binaries in Globular Clusters"
    by Benjamin Giesers
    https://ediss.uni-goettingen.de/handle/21.11130/00-1735-0000-0005-13B4-A
    """

    def __init__(self, model_x: np.ndarray, model_y: np.ndarray, min_bunch_size: int = 3, max_ptp: float = 0.001,
                 threshold: float = 0.9995, iterations: int = 4, bin_size: float = 100,
                 continuum_points: np.ndarray = None, to_mask: List[Tuple[float, float]] = None,
                 standard_mask: bool = True, *args, **kwargs):
        """Normalize spectrum with the help of a fitted model (the model could also be the spectrum itself!)

        It is recommended to run this normalization twice, first with a low-degree polynomial and then with a
        spline. Alternatively, the first iteration can be done using another method like MaxBin.


        Args:
            model_x: Array of model wavelength points
            model_y: Array of model flux points.
            min_bunch_size: The minimum number of coherent flux points within a continuum bunch.
            max_ptp: The maximum peak to peak deviation of flux points within a continuum bunch.
            threshold: Threshold to filter wrong (local minima) continuum bunches (comparison with neighbor bunches).
            iterations: Filter iterations to find wrong continuum bunches.
            bin_size: Wavelength bins to calculate (spline fit) interpolated continuum.
            continuum_points: Definitive mask of continuum points in spectrum (wavelength). Will be ignored if None.
            to_mask: Wavelength regions [min, max] to mask, where continuum will not be detected. If None standard set is taken.
            standard_mask: Should a standard mask for broad lines be applied?
        """
        Continuum.__init__(self, *args, **kwargs)

        # store
        self._min_bunch_size = min_bunch_size
        self._max_ptp = max_ptp
        self._threshold = threshold
        self._iterations = iterations
        self._bin_size = bin_size

        # mask tellurics and wings of deep absorptions
        if to_mask is None:
            to_mask = []
        if standard_mask is True:
            to_mask += [[4000, 4768], [4810, 4930], [5763, 5775], [5877, 5898],
                        [6266, 6277], [6515, 6640], [9333, 10000]]
        self._mask = np.ones(len(model_x), dtype=bool)
        for m in to_mask:
            self._mask &= (model_x < m[0]) | (model_x > m[1])

        # get continuum points and continuum in model spectrum
        cont, bunches, bunch_index = self._determine_continuum_bunches(model_y)
        self._cont = self._get_filtered_continuum(cont, model_y, bunches, bunch_index)

        # if continuum_points is set, use them definitely
        if continuum_points is not None:
            self._cont |= continuum_points

    def __call__(self, x: np.ndarray, y: np.ndarray, valid: np.ndarray = None) -> np.ndarray:
        """Calculate the continuum.

        Args:
            x: x array for continuum to fit.
            y: y array for continuum to fit.
            valid: Valid pixels mask.

        Returns:
            Array containing calculated continuum.
        """

        # create continuum spectrum
        cont_x = x[self._cont & self._mask & ~np.isnan(y)]
        cont_y = y[self._cont & self._mask & ~np.isnan(y)]

        # bin continuum points in spectrum
        bin_wave = []
        bin_flux = []
        for ind in range(int((np.max(x) - np.min(x)) / self._bin_size) + 1):
            slc = (cont_x > np.min(x) + ind * self._bin_size) & (cont_x < np.min(x) + (ind + 1) * self._bin_size)
            if len(cont_x[slc]) > 2:
                bin_wave.append(np.mean(cont_x[slc]))
                bin_flux.append(np.median(cont_y[slc]))

        # deal with boundary conditions
        bin_wave.append(np.max(x))
        bin_flux.append(bin_flux[-1])
        bin_wave.insert(0, np.min(x))
        bin_flux.insert(0, bin_flux[0])

        # fit continuum
        return self._fit_poly(bin_wave, bin_flux, xout=x)

    def _determine_continuum_bunches(self, y) -> Tuple[np.ndarray, List[int], np.ndarray]:
        """Determine bunches of continuum in given spectrum.

        Args:
            y: Array of spectrum flux points

        Returns:
            Tuple containing:
                - Mask of continuum points
                - List of bunch names
                - Mask with bunch names
        """

        continuum = []
        bunches = []
        bunch_index = []
        start = 0
        # go through spectrum
        while start < len(y):
            # reset peak to peak value and size of bunch
            ptp = 0
            size = 0
            # create bunch until max_ptp exceeded or end of spectrum reached
            while ptp < self._max_ptp and start + size < len(y):
                size += 1
                ptp = np.ptp(y[start:start + size] / np.median(y))

            # reduce size by 1, cause last point is not continuum
            size -= 1

            # fulfilling continuum criteria?
            if size >= self._min_bunch_size:
                continuum += list(np.ones(size, dtype=bool))

                # store new bunch
                new_bunch = len(bunches) + 1
                bunches.append(new_bunch)

                bunch_index += list(np.ones(size, dtype=int) * new_bunch)
                start += size
            else:
                continuum += [False]

                bunch_index += [0]
                start += 1

        return np.array(continuum), bunches, np.array(bunch_index)

    def _get_filtered_continuum_bunches(self, y, bunches, bunch_index):
        # store first bunch, cause we can only start filtering with second
        filtered_bunches = [bunches[0]]

        # calculate median of all bunches spectrum flux
        median = {}
        for bunch in bunches:
            median[bunch] = np.median(y[bunch_index == bunch])

        # compare each bunch with neighbors, starting with second and ending with second last
        for i in range(1, len(bunches) - 1):
            if median[bunches[i]] > self._threshold * median[bunches[i - 1]] or \
                    median[bunches[i]] > self._threshold * median[bunches[i + 1]]:
                filtered_bunches.append(bunches[i])

        # store last bunch
        filtered_bunches.append(bunches[-1])

        return filtered_bunches

    def _get_filtered_continuum(self, continuum, spectrum_flux, bunches, bunch_index):
        # iterate to identify wrong continuum bunches
        for _ in range(self._iterations):
            bunches = self._get_filtered_continuum_bunches(spectrum_flux, bunches, bunch_index)

        # remove wrong continuum bunches from continuum
        for i, bunch in enumerate(bunch_index):
            if bunch not in bunches:
                continuum[i] = False

        return continuum


__all__ = ['Continuum', 'MaximumBin', 'Regions', 'SigmaClipping', 'PeakToPeak']



import numpy as np

from ..object import spexxyObject
from ..data import Spectrum


class Mask(spexxyObject):
    """Mask is the base class for all objects that can create a good pixel mask for spectra."""

    def __init__(self, *args, **kwargs):
        """Initialize a new mask."""
        spexxyObject.__init__(self, *args, **kwargs)

    def __call__(self, spectrum: Spectrum, filename: str) -> np.ndarray:
        """Creates a new mask for a spectrum.

        Args:
            spectrum: Spectrum to create mask for.
            filename: Name of file containing spectrum to create mask for.

        Returns:
            Boolean array containing good pixel mask for given spectrum.

        """
        raise NotImplementedError


__all__ = ['Mask']



import os
import logging
from astropy.io import fits
import numpy as np

from .weight import Weight
from ..data import Spectrum


class WeightFromSigma(Weight):
    """Reads the SIGMA extension from the given file and creates weights from it as 1/SIGMA.

    This class, when called, loads the SIGMA extension from the given file and returns the weights as 1/SIGMA.
    """

    def __init__(self, squared: bool = False, *args, **kwargs):
        """Initializes a new weight.

        Args:
            squared: Return 1/SIGMA**2 instead of 1/SIGMA.
        """
        Weight.__init__(self, *args, **kwargs)
        self._squared = squared

    def __call__(self, spectrum: Spectrum, filename: str) -> np.ndarray:
        """Creates a new weight for a spectrum from its SIGMA extension.

        Args:
            spectrum: Spectrum to create weight for.
            filename: Name of file containing spectrum to create weight for.

        Returns:
            Array containing weight for given spectrum.
        """

        self.log.info('Creating weights from SIGMA extension...')

        # load SIGMA
        sigma = fits.getdata(filename, 'SIGMA')

        # calculate weights
        weights = 1. / sigma

        # square it?
        if self._squared:
            weights = weights * weights

        # finished
        return weights


__all__ = ['WeightFromSigma']



import logging
import argparse


log = logging.getLogger(__name__)


def add_parser(subparsers):
    # init parser
    parser = subparsers.add_parser('create', help='Create a new files grid',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('root', type=str, help='Path to create grid from', default='.')
    parser.add_argument('-o', '--output', type=str, help='Output database file', default='grid.csv')
    parser.add_argument('-p', '--pattern', type=str, help='Filename pattern', default='**/*.fits')
    parser.add_argument('--from-filename', type=str, help='Parse parameters from the filenames with the given '
                                                          'regular expression')
    # argparse wrapper for create_grid
    def run(args):
        create_grid(**vars(args))
    parser.set_defaults(func=run)


def create_grid(root, output, pattern, from_filename, **kwargs):
    import os
    import glob
    import re

    # easiest way to list all files, is to remember current directory, go to root, glob, and go back
    cwd = os.getcwd()
    os.chdir(root)
    files = sorted(glob.glob(pattern, recursive=True))
    os.chdir(cwd)

    # how do we get the parameters?
    if from_filename is not None:
        # create regexp
        regexp = re.compile(from_filename)

        # get parameters
        parameters = list(regexp.groupindex.keys())

        # open file
        with open(output, 'w') as csv:
            csv.write('Filename,%s\n' % ','.join(parameters))

            # loop files
            for filename in files:
                # do matching
                m = regexp.search(filename)
                if m is None:
                    continue

                # get values by looping values of groupindex and get values of corresponding groups
                # this works in the correct order, because Python guarantees, that keys() and values() on a dict
                # have the same order
                # the +0 avoids negative zeros
                values = [0. if m.group(idx) is None else float(m.group(idx)) + 0. for idx in regexp.groupindex.values()]

                # write to csv
                csv.write('%s,%s\n' % (filename, ','.join([str(v) for v in values])))



import os
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.io import fits
from astropy.time import Time
import astropy.units as u

from .init import Init
from ..component import Component


class InitFromVhelio(Init):
    """Initializes a component from the heliocentric correction calculated for RA/Dec given in the FITS headers.

    This class, when called, initializes a single parameter (the radial velocity, default to "v") of the given
    component with the heliocentric correction calculated from RA/DEC/DATE-OBS in the given file.
    """

    def __init__(self, negative: bool = False, parameter: str = 'v', scale: str = 'utc', obs: str = 'paranal',
                 kind: str = 'barycentric', *args, **kwargs):
        """Initializes a new Init object.

        Args:
            negative: If True, uses the negative of the calculated correction.
            parameter: Name of the parameter in the component to set.
            scale: Time scale to use for DATE-OBS.
            obs: Observatory to use for calculating correction.
            kind: Either barycentric or heliocentric.
        """
        Init.__init__(self, *args, **kwargs)
        self._negative = negative
        self._parameter = parameter
        self._scale = scale
        self._obs = EarthLocation.of_site(obs)
        self._kind = kind

    def __call__(self, cmp: Component, filename: str):
        """Initializes the radial velocity parameter of a given component to the (negative) heliocentric correction.

        Args:
            cmp: Component to initialize.
            filename: Name of file containing RA/DEC/DATE-OBS in FITS header to calculate correction from.
        """
        # file doesn't exist?
        if not os.path.exists(filename):
            self.log.error('Give file %s for extracting coordinates does not exist, skipping.', filename)
            return

        # get fits header
        header = fits.getheader(filename)

        # got coordinates in header?
        if 'RA' in header and 'DEC' in header and 'DATE-OBS' in header:
            # construct SkyCoord and time
            sc = SkyCoord(ra=header['RA'] * u.deg, dec=header['DEC'] * u.deg)
            time = Time(header['DATE-OBS'], scale=self._scale)

            # calculate correction
            vhelio = sc.radial_velocity_correction(self._kind, obstime=time, location=self._obs).to(u.km/u.s).value
            if self._negative:
                vhelio *= -1

            # set it
            self.log.info('Initializing "%s" of component "%s" to %sv_helio=%f...',
                          self._parameter, cmp.prefix, '-' if self._negative else '', vhelio)
            cmp[self._parameter] = vhelio


__all__ = ['InitFromVhelio']



import os
import pkgutil
import importlib


def add_tree_node(parser, package, help, aliases=None):
    # package name
    name = package.__name__
    cmd = name[name.rfind('.')+1:]

    # aliases
    if aliases is None:
        aliases = []

    # add parser
    p = parser.add_parser(cmd, help=help, aliases=aliases)
    sp = p.add_subparsers(help=help)

    # list modules in gcdb.methods.fits
    pkgpath = os.path.dirname(package.__file__)
    modules = [name for _, name, _ in pkgutil.iter_modules([pkgpath])]

    # loop modules
    for m in modules:
        # import module
        mod = importlib.import_module(name + '.' + m)

        # add subparser
        mod.add_parser(sp)



import logging
import argparse
import os
import shutil
from typing import List

from spexxy.grid import Grid

log = logging.getLogger(__name__)


def add_parser(subparsers):
    # init parser
    parser = subparsers.add_parser('combine', help='Combine multiple grids into one',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('output', type=str, help='Output path')
    parser.add_argument('sources', type=str, nargs='+', help='Input CSV files')

    # argparse wrapper for combine_grid
    def run(args):
        combine_grid(**vars(args))
    parser.set_defaults(func=run)


def combine_grid(output: str, sources: List[str], **kwargs):
    # load first grid
    grid = Grid.load(sources[0])

    # outdir and grid file
    if not os.path.exists(output):
        os.makedirs(output)
    with open(os.path.join(output, 'grid.csv'), 'w') as f:
        f.write('Filename,' + ','.join(grid.axis_names()) + '\n')

    # store existing parameters
    existing_params = []

    # loop all input grids
    for src in sources:
        # load grid
        grid = Grid.load(src)

        # only filegrids supported
        if not hasattr(grid, 'filename'):
            print('Only filegrids supported.')
            continue

        # get all params
        all_params = grid.all()

        # loop params
        for params in all_params:
            # does it exist?
            if params in existing_params:
                continue

            # get name and directory of output file
            filename = grid.filename(params)
            basepath = os.path.relpath(filename, os.path.dirname(src))
            outfile = os.path.join(output, basepath)
            outpath = os.path.dirname(outfile)

            # does dir exist?
            if not os.path.exists(outpath):
                os.makedirs(outpath)

            # copy
            log.info("  Copying to {0:s}...".format(outfile))
            shutil.copyfile(filename, outfile)

            # add to CSV
            with open(os.path.join(output, 'grid.csv'), 'a') as csv:
                csv.write(filename + ',' + ','.join([str(p) for p in params]) + '\n')

            # add params
            existing_params.append(params)



import math
import os
from fnmatch import fnmatch
from typing import List

import numpy as np
from scipy.integrate import trapz
from spexxy.data.spectrum import Spectrum, SpectrumAscii


class Filter(object):
    """ A filter class than can be applied to spectra """

    def __init__(self, filter_name: str = None, file_extension: str = '.txt', path: str = '$SPEXXYPATH/filters'):
        """Initializes a filter object with the given filter.

        Names for files that contain filter data are build as <path>/filters/<filter_name>.<extension>.

        Args:
            filter_name: Name of filter data to load without extension
            file_extension: Extension for file.
            path: Path the file is located in.
        """

        # init
        self.filter_name = filter_name
        self._throughput = None
        self._vega = None
        self._vega_throughput = None

        # if no filter name is given, we're finished here
        if filter_name is None:
            return

        # is filter_name a filename?
        if os.path.exists(filter_name):
            # just set it
            self._filter_filename = filter_name
        else:
            # build filename
            self._filter_filename = os.path.join(path, filter_name + file_extension)

        # load filter
        self._throughput = SpectrumAscii(self._filter_filename, separator=None)

    @property
    def filename(self):
        """Filename for filter"""
        return self._filter_filename

    @property
    def wave(self):
        """Wavelength array the filter is defined on"""
        return self._throughput.wave

    @property
    def throughput(self):
        """Filter response curve"""
        return self._throughput.flux

    def copy(self):
        """Copy filter

        Returns:
            Copy of this filter
        """

        # create new filter
        f = Filter()

        # copy data
        f.filter_name = self.filter_name
        f._filter_filename = self._filter_filename
        f._throughput = self._throughput.copy() if self.throughput is not None else None
        f._vega = self._vega.copy() if self._vega is not None else None
        f._vega_throughput = self._vega_throughput.copy() if self._vega_throughput is not None else None

        # return it
        return f

    @staticmethod
    def list(pattern: str = '*', file_extension: str = '.txt', path: str = '$SPEXXYPATH/filters') -> List['Filter']:
        """Load all filters in a given directory that match the pattern and file extension.

        Args:
            pattern: Pattern for filters, e.g. 'Johnson/*' or 'hst/acs/*'.
            file_extension: File extension of filter files in given directory.
            path: Root path to start search in.

        Returns:
            List of filters that match the criteria.
        """

        # expand path
        p = os.path.expandvars(path)

        # walk path
        filters = []
        for root, dirs, files in os.walk(p):
            for file in sorted(files):
                filename = os.path.splitext(os.path.join(root, file)[len(p) + 1:])[0]
                if fnmatch(filename, pattern):
                    filters.append(Filter(filename, file_extension=file_extension, path=path))
        return filters
        
    def resample(self, inplace: bool = True, *args, **kwargs):
        """Resamples filter to other wavelength grid.

        Args:
            inplace: If True, this filter is changed, otherwise a new one is returned.
            args, kwargs: Forwarded to Spectrum's resample method.

        Returns:
            If inplace is True, a new Filter with the resampled response curve.
        """

        # do linear interpolation by default
        if 'linear' not in kwargs:
            kwargs['linear'] = True

        # fill missing values with zeros
        kwargs['fill_value'] = 0.

        # delegate resampling to Spectrum class
        if inplace:
            self._throughput = self._throughput.resample(*args, **kwargs)
        else:
            # create new filter
            f = self.copy()
            f._throughput = f._throughput.resample(*args, **kwargs)
            return f

    def integrate(self, spec: Spectrum) -> float:
        """Apply the filter to a given spectrum and integrate it.

        Args:
            spec: The spectrum to integrate together with the filter.

        Returns:
            Integrated flux of the given spectrum after applying the filter.
        """

        # lambda mode
        if spec.wave_mode != Spectrum.Mode.LAMBDA:
            raise ValueError('Spectrum must be in LAMBDA mode.')

        # wavelength grid okay?
        if not np.array_equal(spec.wave, self._throughput.wave):
            raise ValueError('Please resample filter before applying it.')

        # apply filter
        fs = spec.flux * self._throughput.flux

        # integrate
        return trapz(fs, spec.wave) / trapz(self._throughput.flux, spec.wave)

    def stmag(self, spec: Spectrum) -> float:
        """Calculate ST magnitude of given spectrum.

        Args:
            spec: The spectrum to get ST magnitude for.

        Returns:
            ST magnitude of the given spectrum in this filter.
        """

        # lambda mode
        if spec.wave_mode != Spectrum.Mode.LAMBDA:
            raise ValueError('Spectrum must be in LAMBDA mode.')

        # convolve with filter and integrate
        integral = self.integrate(spec)

        # ST magnitude
        return -2.5 * math.log10(integral) - 21.1

        # calculate stmag
        return self.stmag(spec.wave, spec.flux)

    def _load_vega(self):
        """Load VEGA spectrum."""
        if self._vega is None or self._vega_throughput is None:
            # get path to vega.sed
            filename = os.path.join(os.path.dirname(__file__), 'vega.sed')

            # no vega given? use default
            self._vega = SpectrumAscii(filename, separator=None)

            # create filter for vega
            tmp = SpectrumAscii(self._filter_filename, separator=None)
            self._vega_throughput = tmp.resample(spec=self._vega, fill_value=0., linear=True)

    def vegamag(self, spec: Spectrum) -> float:
        """Calculate Vega magnitude of given spectrum.

        Args:
            spec: The spectrum to get Vega magnitude for.

        Returns:
            Vega magnitude of the given spectrum in this filter.
        """

        # load vega spectrum
        self._load_vega()

        # lambda mode
        if spec.wave_mode != Spectrum.Mode.LAMBDA:
            raise ValueError('Spectrum must be in LAMBDA mode.')

        # no vega given?
        if self._vega is None:
            raise ValueError('No Vega spectrum given.')

        # wavelength grid okay?
        if not np.array_equal(spec.wave, self._throughput.wave):
            raise ValueError('Please resample filter before applying it.')

        # if no filter is given, use full spectrum
        if self.filter_name is None:
            # whole spectrum
            flux1 = trapz(spec.flux, spec.wave)
            flux2 = trapz(self._vega.flux, self._vega.wave)

        else:
            # multiply spectra with filter
            filter_spec = spec.flux * self._throughput.flux
            filter_vega = self._vega.flux * self._vega_throughput.flux

            # integrate
            w = ~np.isnan(filter_spec)
            flux1 = trapz(filter_spec[w], spec.wave[w]) / trapz(self._throughput.flux[w], spec.wave[w])
            w = ~np.isnan(filter_vega)
            flux2 = trapz(filter_vega[w], self._vega.wave[w]) / trapz(self._vega_throughput.flux[w], self._vega.wave[w])

        # check
        if flux2 == 0:
            raise ValueError('Integrated flux for Vega spectrum is zero.')

        # calculate Vega magnitude
        mag = -2.5 * math.log10(flux1 / flux2)
        return mag


__all__ = ['Filter']



import numpy as np

from .mask import Mask
from ..data import Spectrum


class MaskEnds(Mask):
    """Masks the ends of a spectrum.

    This class, when called, creates a mask that masks the N pixels at each end of the given spectrum.
    """

    def __init__(self, npixels=10, *args, **kwargs):
        """Initializes a new mask masking the ends of a spectrum.

        Args:
            npixels: Number of pixels to mask at each end of the spectrum.
        """
        Mask.__init__(self, *args, **kwargs)
        self._npixels = npixels

        # npixels cannot be negative
        if npixels < 0:
            raise ValueError('Number of pixels to mask at each end of the spectrum cannot be negative.')

    def __call__(self, spectrum: Spectrum, filename: str = None) -> np.ndarray:
        """Creates a new mask for the given spectrum masking only the N pixels at each end.

        Args:
            spectrum: Spectrum to create mask for.
            filename: Name of file containing spectrum to create mask for (unused).

        Returns;
            Boolean array containing good pixel mask for given spectrum.
        """

        # create empty mask
        mask = np.ones((len(spectrum.wave)), dtype=bool)

        # if npixels is zero, just return mask
        if self._npixels == 0:
            return mask

        # mask ends and return mask
        mask[:self._npixels] = False
        mask[-self._npixels:] = False
        self.log.info('Masked %d pixels at both ends of the spectrum.', self._npixels)
        return mask


__all__ = ['MaskEnds']



import os
from typing import Callable, Union

from .component import Component
from ..data import LOSVD, Spectrum, LSF
from ..object import create_object


class SpectrumComponent(Component):
    """SpectrumComponent is the base Component class for all components that deal with spectra."""

    def __init__(self, name: str, losvd_hermite: bool = False, vac_to_air: bool = False,
                 lsf: Union[LSF, str, dict] = None, *args, **kwargs):
        """Initializes a new SpectrumComponent.

        Args:
            name: Name of new component
            losvd_hermite: Whether or not Hermite polynomials should be used for the LOSVD
            vac_to_air: If True, vac_to_air() is called on spectra returned from the model_func
            lsf: LSF to apply to spectrum
        """
        Component.__init__(self, name, *args, **kwargs)
        self._vac_to_air = vac_to_air
        self._losvd_hermite = losvd_hermite

        # add losvd parameters
        self.set('v', min=-2000., max=2000., value=1.)
        self.set('sig', min=0., max=500., value=10.)
        if self._losvd_hermite:
            self.set('h3', min=-0.3, max=0.3, value=0.)
            self.set('h4', min=-0.3, max=0.3, value=0.)
            self.set('h5', min=-0.3, max=0.3, value=0.)
            self.set('h6', min=-0.3, max=0.3, value=0.)

        # get lsf
        if isinstance(lsf, LSF):
            self._lsf = lsf
        elif isinstance(lsf, str):
            self._lsf = LSF.load(lsf)
        elif isinstance(lsf, dict):
            self._lsf = create_object(lsf)
        else:
            self._lsf = None

    def _model_func(self):
        """Return actual model."""
        raise NotImplementedError

    def __call__(self, **kwargs):
        """Model function that creates a spectrum with the given parameters and shift/convolve it using the given LOSVD.

        Args:
            kwargs: Values to overwrite

        Returns:
            The model from the component
        """

        # overwrite any?
        for key, val in kwargs.items():
            if key in self.parameters:
                self.set(key, value=val)

        # get LOSVD parameters
        losvd_params = ['v', 'sig'] + (['h3', 'h4', 'h5', 'h6'] if self._losvd_hermite else [])
        losvd = [self[p] for p in losvd_params]

        # get  model
        model = self._model_func()

        # apply LSF
        if self._lsf is not None:
            self._lsf.resample(model)
            model = self._lsf(model)

        # apply losvd
        self._apply_losvd(model, losvd)

        # vac2air
        if self._vac_to_air:
            model.vac_to_air()

        # return result
        return model

    @staticmethod
    def _apply_losvd(model, losvd):
        """Apply LOSVD with the given parameters to the given model.

        WARNING: We will NOT FIT line broadening, if model spectra are in LAMBDA mode!

        Args:
            model: Model to apply LOSVD to.
            losvd: LOSVD parameters (v, sig, <h3, h4, h5, h6>)
        """
        if losvd[1] < 1e-5 or model.wave_mode == Spectrum.Mode.LAMBDA:
            # in LAMBDA mode, no LOSVD is supported
            model.redshift(losvd[0])
        else:
            # full LOSVD for sig>0 and LOG mode
            if model.wave_step == 0:
                # resample to const, if necessary
                model = model.resample_const()
            # apply losvd
            losvd = LOSVD(losvd)
            model.flux = losvd(model)


__all__ = ['SpectrumComponent']



import numpy as np
from typing import List, Tuple, Any

from ..object import spexxyObject


class GridAxis(object):
    """Description of a single axis in a grid."""

    def __init__(self, name: str, values: List = None, min: float = None, max: float = None, initial: float = None):
        """Initialize a new grid axis.

        Args:
            name: Name of new axis.
            values: List of all possible values for this axis.
            min: Minimum value for this axis.
            max: Maximum value for this axis.
            initial: Initial guess for this axis.
        """

        # remember name
        self.name = name

        # store values and get min/max
        self.values = None if values is None else sorted(values)
        if (min is None or max is None) and values is not None:
            self.min = self.values[0]
            self.max = self.values[-1]
        else:
            self.min = min
            self.max = max

        # store initial value - if none is given, take mean of values
        self.initial = np.mean(self.values) if initial is None else initial

    def neighbour(self, value: float, distance: int=1) -> float:
        """Finds a neighbour in this axis for the given value in the given distance.

        Args:
            value: Value to find neighbour for.
            distance: Distance in which to find neighbour
                >0:  Find larger neighbours, i.e. 0 next larger value, 1 the one after that, etc
                <=0:  Find smaller neighbours, i.e. 0 next smaller value (or value itself), -1 the before that, etc

        Returns:
            Value on grid in the given distance to the given value. If given value is on grid, distance is
            counted from that value.

        Raises:
            KeyError: If no neighbour has been found.
        """

        # loop all values
        for i in range(len(self.values)):
            # found value?
            if self.values[i] <= value < self.values[i + 1]:
                # index of neighour
                ii = i + distance
                # does it exist?
                if 0 <= ii < len(self.values):
                    return self.values[ii]

        else:
            # nothing found
            raise KeyError('No neighbour found.')


class Grid(spexxyObject):
    """Base class for all grids in spexxy."""

    def __init__(self, axes: List[GridAxis], *args, **kwargs):
        """Initialize a new Grid.

        Args:
            axes: List of axes for this grid.
        """
        spexxyObject.__init__(self, *args, **kwargs)

        # store axes
        self._axes = axes

    def num_axes(self) -> int:
        """Returns number of axes.

        Returns:
            Number of axes in grid.
        """
        return len(self._axes)

    def axes(self) -> List[GridAxis]:
        """Returns information about the axes.

        Returns:
            List of GridAxis objects describing the grid's axes.
        """
        return self._axes

    def axis_values(self, axis: int) -> List[float]:
        """Returns all possible values for the given axis.

        Args:
            axis: Index of axis to return values for.

        Returns:
            All possible values for the given axis.
        """
        return self._axes[axis].values

    def axis_name(self, axis: int) -> str:
        """Returns name of given axis.

        Args:
            axis: Index of axis to return values for.

        Returns:
            Name of given axis.
        """
        return self._axes[axis].name

    def axis_names(self) -> List[str]:
        """Returns names of all axis.

        Returns:
            Names of all axis.
        """
        return [axis.name for axis in self._axes]

    def all(self) -> List[Tuple]:
        """Return all possible parameter combinations.

        Returns:
            All possible parameter combinations.
        """
        raise NotImplementedError

    def __contains__(self, params: Tuple) -> bool:
        """Checks, whether the grid contains a given parameter set.

        Args:
            params: Parameter set to check.

        Returns:
            Whether or not the given parameter set exists in the grid.
        """
        raise NotImplementedError

    def __call__(self, params: Tuple) -> Any:
        """Fetches the value for the given parameter set

        Args:
            params: Parameter set to catch value for.

        Returns:
            Grid value at given position.
        """
        raise NotImplementedError

    def create_array(self) -> Any:
        """In case the values provided by this grid are of an array type, this method creates an empty array that
        can be filled.

        Returns:
            Empty array element of same type as values in the grid.
        """

        # fetch first item from grid
        data = self(next(iter(self.all())))

        # zero and return it
        data *= 0.
        return data

    def neighbour(self, params: Tuple, axis: int, distance: int=1, must_exist: bool = False) -> Tuple:
        """Finds a neighbour on the given axis for the given value in the given distance.

        Args:
            params: Parameter tuple to search neighbour from.
            axis: Axis to search for
            distance: Distance in which to find neighbour.
                >0:  Find larger neighbours, i.e. 0 next larger value, 1 the one after that, etc
                <=0:  Find smaller neighbours, i.e. 0 next smaller value (or value itself), -1 the before that, etc
            must_exist: Grid point with new parameter set must actually exist.

        Returns:
            New parameter tuple with neighbour on the given axis.
        """

        # create new tuple
        p = list(params)

        # if we don't enfore the new point to exist, it's easy
        if not must_exist:
            # find neighbour in axis
            value = self._axes[axis].neighbour(params[axis], distance=distance)
            if value is None:
                return None

            # create new tuple
            p[axis] = value
            return tuple(p)

        # grid point does not exist, get all axis values
        axis_values = np.array(self._axes[axis].values)

        # lower or higher values?
        if distance <= 0:
            # smaller values
            axis_values = sorted(axis_values[axis_values < params[axis]], reverse=True)
            # define steps to take
            steps = abs(distance)
        else:
            # larger values
            axis_values = sorted(axis_values[axis_values > params[axis]], reverse=False)
            # define steps to take, subtract 1, so that 1 (first larger) is index 0
            steps = distance - 1

        # find parameter set that exists
        for v in axis_values:
            # set value
            p[axis] = v

            # check it
            if self.__contains__(tuple(p)):
                # exists, does distance match, i.e. steps=0?
                if steps == 0:
                    # return tuple
                    return tuple(p)
                else:
                    # otherwise, reduce steps and go on
                    steps -= 1
        else:
            # nothing found
            return None

    def nearest(self, params, scales=None) -> Tuple:
        """Finds the nearest point within the grid.

        Calculates distance between given params and all points p in the grid as the sum over all elements in
        the parameter tuple of ((params - grid_point) * scales)**2.

        Args:
            params: Parameters to find nearest point to.
            scales: If given, scales dimensions.

        Returns:
            Parameters for nearest point in grid.
        """

        # check scales
        num_axes = self.num_axes()
        if scales and len(scales) != num_axes:
            raise ValueError('Invalid number of scales, must be identical '
                             'to number of axes.')
        scales = np.array(scales) if scales is not None else np.ones((num_axes))

        # check params
        if len(params) != num_axes:
            raise ValueError('Number of parameters must be equal to '
                             'number of axes.')

        # get all parameters values and convert them to numpy arrays
        all_params = self.all()

        # calc squared distances and scale:
        # - distance: params - p
        # - scale:    * scales
        # - squared:  **2
        # - sum up:   np.sum
        # There is no need to take the square root, since we're just
        # comparing values.
        dist = [np.sum(((params - np.array(p)) * scales)**2) for p in all_params]

        # get best match
        return all_params[np.argmin(dist)]

    @staticmethod
    def load(filename: str, **kwargs) -> 'Grid':
        """Tries to identify type of grid and loads it.

        Args:
            filename: Name of file to load.
        """

        # in this naive implementation, we just check file type
        if filename.endswith('.csv'):
            # might be a files grid
            from .files import FilesGrid
            return FilesGrid(filename, **kwargs)

        elif filename.endswith('.dat'):
            # might be a ferre grid
            from .ferre import FerreGrid
            return FerreGrid(filename)

        elif filename.endswith('.fits'):
            # FITS grid
            from .fits import FitsGrid
            return FitsGrid(filename, **kwargs)

        else:
            raise ValueError('Unknown format')


__all__ = ['GridAxis', 'Grid']
